/* -*- tab-width: 4 -*- */

#include "global.h"
#include "elkkernel.h"

#include <errno.h>
#if defined(MACOS) && !defined(MACOSX)
#  include <unix.h>
#else
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <errno.h>
#endif

#ifndef MAX_STACK_SIZE
#  include <sys/resource.h>
#endif

#ifdef FIND_AOUT
#  ifdef HAVE_UNISTD_H
#	include <unistd.h>
#  else
#	include <sys/file.h>
#  endif
#endif

CHAR *stkbase;
int Stack_Grows_Down;
int Max_Stack;
int Interpreter_Initialized;
int GC_Debug = 0;
int Case_Insensitive;
int Verb_Load, Verb_Init;

CHAR **Argv;
int Argc, First_Arg;

void Get_Stack_Limit (void);
void Usage (void);
void Init_Everything (void);

#ifdef FIND_AOUT
CHAR *A_Out_Name;
CHAR *Find_Executable();
#endif

#if defined(CAN_LOAD_OBJ) || defined(INIT_OBJECTS)
SYMTAB *The_Symbols;
#endif

void Exit_Handler (void) {
#if defined(CAN_LOAD_OBJ) || defined(INIT_OBJECTS)
	Call_Finalizers ();
#endif
#ifdef CAN_LOAD_OBJ
	Finit_Load ();
#endif
}

#ifndef ATEXIT
/* Hack: __GNUC_MINOR__ was introduced together with __attribute__ */
#ifdef __GNUC_MINOR__
extern void _exit() __attribute__ ((noreturn));
#endif
#endif

#ifdef CAN_DUMP
int Was_Dumped;
CHAR *Brk_On_Dump;
#endif


/* dump currently does not work for applications using Elk_Init().
 * The reason is that in this case the INITIAL_STK_OFFSET which
 * compensates for differences in argv[] in the original/dumped a.out
 * is not in effect (see comment below).
 * This cannot be fixed without changing Elk_Init() and its use in
 * an incompatible way.
 */
void Check_If_Dump_Works (void) {
#ifdef NOMAIN
	Primitive_Error (x_("not yet supported for standalone applications"));
#endif
}


#ifdef NOMAIN

void Elk_Init (int ac, CHAR **av, int init_objects, CHAR *toplevel) {

#else

main (int ac, CHAR **av) {

#endif

/* To avoid that the stack copying code overwrites argv if a dumped
 * copy of the interpreter is invoked with more arguments than the
 * original a.out, move the stack base INITIAL_STK_OFFSET bytes down.
 * The call to bzero() is there to prevent the optimizer from removing
 * the array.
 */
#ifdef CAN_DUMP
	CHAR unused[INITIAL_STK_OFFSET];
#endif
	CHAR *initfile, *loadfile = 0, *loadpath = 0;
	int debug = 0, heap = HEAP_SIZE, leng;
	ELKObject file;
	struct stat st;
	CHAR foo = 0;
#ifdef NOMAIN
#  define foo (av[0][0])
#endif

#ifdef CAN_DUMP
	bzero (unused, 1);  /* see comment above */
#endif
	if (ac == 0) {
		av[0] = x_("Elk"); ac = 1;
	}
	Get_Stack_Limit ();

#ifdef FIND_AOUT
	A_Out_Name = Find_Executable (av[0]);
#endif

	Argc = ac; Argv = av;
	First_Arg = 1;
#ifdef CAN_DUMP
	if (Was_Dumped) {
		/* Check if beginning of stack has moved by a large amount.
		 * This is the case, for instance, on a Sun-4m when the
		 * interpreter was dumped on a Sun-4c and vice versa.
		 */
		if (abs (stkbase - &foo) > INITIAL_STK_OFFSET) {
			ttyputerr(x_("Can't restart dumped interpreter from a different machine architecture"));
			ttyputerr(x_("   (Stack delta = %d bytes)."), stkbase - &foo);
			exit (1);
		}
		/* Check if program break must be reset.
		*/
		if (Brk_On_Dump && (CHAR *)brk (Brk_On_Dump) == (CHAR *)-1) {
			perror (x_("brk")); exit (1);
		}
#if defined(HP9K) && defined(CAN_DUMP) && defined(HPSHLIB)
		Restore_Shared_Data ();
#endif
#ifdef GENERATIONAL_GC
		Generational_GC_Reinitialize ();
#endif
		Loader_Input = 0;
		Install_Intr_Handler ();
		(void)Funcall_Control_Point (Dump_Control_Point, Arg_True, 0);
		/*NOTREACHED*/
	}
#endif

	for ( ; First_Arg < ac; First_Arg++) {
		if (estrcmp (av[First_Arg], x_("-g")) == 0) {
			debug = 1;
		} else if (estrcmp (av[First_Arg], x_("-i")) == 0) {
			Case_Insensitive = 1;
		} else if (estrcmp (av[First_Arg], x_("-v")) == 0) {
			if (++First_Arg == ac)
				Usage ();
			if (estrcmp (av[First_Arg], x_("load")) == 0)
				Verb_Load = 1;
			else if (estrcmp (av[First_Arg], x_("init")) == 0)
				Verb_Init = 1;
			else Usage ();
		} else if (estrcmp (av[First_Arg], x_("-h")) == 0) {
			if (++First_Arg == ac)
				Usage ();
			if ((heap = eatoi (av[First_Arg])) <= 0) {
				ttyputerr(x_("Heap size must be a positive number."));
				exit (1);
			}
		} else if (estrcmp (av[First_Arg], x_("-l")) == 0) {
			if (++First_Arg == ac || loadfile)
				Usage ();
			loadfile = av[First_Arg];
		} else if (estrcmp (av[First_Arg], x_("-p")) == 0) {
			if (++First_Arg == ac || loadpath)
				Usage ();
			loadpath = av[First_Arg];
		} else if (estrcmp (av[First_Arg], x_("--")) == 0) {
			First_Arg++;
			break;
		} else if (av[First_Arg][0] == '-') {
			Usage ();
		} else {
			break;
		}
	}

	stkbase = &foo;
	Stack_Grows_Down = Check_Stack_Grows_Down ();
	ALIGN(stkbase);
	Make_Heap (heap);
	Init_Everything ();
#ifdef ATEXIT
	if (atexit (Exit_Handler) != 0)
		Fatal_Error (x_("atexit returned non-zero value"));
#endif
#ifdef INIT_OBJECTS
#ifdef NOMAIN
	if (init_objects) {
		Set_Error_Tag (x_("init-objects"));
		The_Symbols = Open_File_And_Snarf_Symbols (A_Out_Name);
		Call_Initializers (The_Symbols, (CHAR *)0, PR_EXTENSION);
	}
#else
	Set_Error_Tag (x_("init-objects"));
	The_Symbols = Open_File_And_Snarf_Symbols (A_Out_Name);
	Call_Initializers (The_Symbols, (CHAR *)0, PR_CONSTRUCTOR);
	Call_Initializers (The_Symbols, (CHAR *)0, PR_EXTENSION);
#endif
#endif
	if (loadpath || (loadpath = egetenv (LOADPATH_ENV)))
		Init_Loadpath (loadpath);

	/* The following code is sort of a hack.  initscheme.scm should not
	 * be resolved against load-path.  However, the .scm-files may not
	 * have been installed yet (note that the interpreter is already
	 * used in the "make" process).
	 * Solution: if initscheme.scm hasn't been installed yet, do search
	 * the load-path, so that -p can be used.
	 */
	Set_Error_Tag (x_("scheme-init"));
	leng = estrlen (el_libdir) + 5 + sizeof (INITFILE) + 1;
	initfile = Safe_Malloc (leng);
	esnprintf(initfile, leng, x_("%slisp%c%s"), el_libdir, DIRSEP, INITFILE);
#ifdef MACOS
	if (stat (initfile, &st) == -1)
#else
	if (stat (initfile, &st) == -1 && errno == ENOENT)
#endif
		file = Make_String (INITFILE, sizeof(INITFILE)-1);
	else
		file = Make_String (initfile, estrlen (initfile));
	free (initfile);
	(void)General_Load (file, The_Environment);

	Install_Intr_Handler ();

	Set_Error_Tag (x_("top-level"));
#ifdef NOMAIN
	if ((loadfile = toplevel) == 0) {
		Interpreter_Initialized = 1;
		GC_Debug = debug;
		return;
	}
#endif
	if (loadfile == 0)
		loadfile = x_("toplevel.scm");
	file = Make_String (loadfile, estrlen (loadfile));
	Interpreter_Initialized = 1;
	GC_Debug = debug;
	if (loadfile[0] == '-' && loadfile[1] == '\0')
		Load_Source_Port (Standard_Input_Port);
	else
		(void)General_Load (file, The_Environment);
#ifndef NOMAIN
	return 0;
#endif
}

static CHAR *Usage_Msg[] = {
	x_("Options:"),
	x_("   [-l filename]   Load file instead of standard toplevel"),
	x_("   [-l -]		  Load from standard input"),
	x_("   [-h heapsize]   Heap size in KBytes"),
	x_("   [-p loadpath]   Initialize load-path (colon-list of directories)"),
	x_("   [-g]			Enable GC-debugging"),
	x_("   [-i]			Case-insensitive symbols"),
	x_("   [-v type]	   Be verbose.  \"type\" controls what to print:"),
	x_("					  load   linker command when loading object file"),
	x_("					  init   names of extension [f]init functions when called"),
	x_("   [--]			End options and begin arguments"),
	0 };

void Usage (void) {
	CHAR **p;

	ttyputerr(x_("Usage: %s [options] [arguments]"), Argv[0]);
	for (p = Usage_Msg; *p; p++)
		ttyputerr(x_("%s"), *p);
	exit (1);
}

void Init_Everything (void) {
	Init_Type ();
	Init_Cstring ();
	Init_String ();
	Init_Symbol ();
	Init_Env ();
	Init_Error ();
	Init_Exception ();
	Init_Io ();
	Init_Prim ();
	Init_Math ();
	Init_Print ();
	Init_Auto ();
	Init_Heap ();
	Init_Load ();
	Init_Proc ();
	Init_Special ();
	Init_Read ();
	Init_Features ();
	Init_Terminate ();
#ifdef CAN_DUMP
	Init_Dump ();
#endif
}

void Get_Stack_Limit (void) {
#ifdef MAX_STACK_SIZE
	Max_Stack = MAX_STACK_SIZE;
#else
	struct rlimit rl;

	if (getrlimit (RLIMIT_STACK, &rl) == -1) {
		perror (x_("getrlimit"));
		exit (1);
	}
	Max_Stack = rl.rlim_cur;
#endif
	Max_Stack -= STACK_MARGIN;
}

#ifdef FIND_AOUT
Executable (CHAR *fn) {
	struct stat s;

	return stat (fn, &s) != -1 && (s.st_mode & S_IFMT) == S_IFREG
			&& access (fn, X_OK) != -1;
}

CHAR *Find_Executable (CHAR *fn) {
	CHAR *path, *dir;
	static CHAR buf[1025];  /* Can't use Path_Max or Safe_Malloc here */
	register CHAR *p;

	for (p = fn; *p; p++) {
		if (*p == '/') {
			if (Executable (fn))
				return fn;
			else
				Fatal_Error (x_("%s is not executable"), fn);
		}
	}
	if ((path = egetenv (x_("PATH"))) == 0)
		path = x_(":/usr/ucb:/bin:/usr/bin");
	dir = path;
	do {
		p = buf;
		while (*dir && *dir != ':')
			*p++ = *dir++;
		if (*dir)
			++dir;
		if (p > buf)
			*p++ = '/';
		estrcpy (p, fn);
		if (Executable (buf))
			return buf;
	} while (*dir);
	if (dir > path && dir[-1] == ':' && Executable (fn))
		return fn;
	Fatal_Error (x_("cannot find pathname of %s"), fn);
	/*NOTREACHED*/
}
#endif

ELKObject P_Command_Line_Args (void) {
	ELKObject ret, tail;
	register int i;
	GC_Node2;

	ret = tail = P_Make_List (Make_Integer (Argc-First_Arg), Null);
	GC_Link2 (ret, tail);
	for (i = First_Arg; i < Argc; i++, tail = Cdr (tail)) {
		ELKObject a;

		a = Make_String (Argv[i], estrlen (Argv[i]));
		Car (tail) = a;
	}
	GC_Unlink;
	return ret;
}

ELKObject P_Exit (int argc, ELKObject *argv) {
	exit (argc == 0 ? 0 : Get_Unsigned (argv[0]));
	/*NOTREACHED*/
	return(Null);
}

#ifdef WIN32
int index(CHAR *str, CHAR c)
{
		int i;
		for(i=0; str[i] != 0; i++)
				if (str[i] == c) return(i+1);
		return(0);
}
#endif
