/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     *********************************************************************
 */

#include "global.h"
#include "sim.h"
#include "simirsim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "usr.h"
#include "network.h"
#if SIMFSDBWRITE != 0
#  include "ffwAPI.h"
#endif

/* #define DUMPH_HOOK */

/********** Simulation vectors **********/

#define NOSIMVECTOR ((SIMVECTOR *)-1)

/* the meaning of SIMVECTOR->command */
#define VECTORCOMMENT    1		/* a comment in the command file */
#define VECTORL          2		/* the "l" command (set signal low) */
#define VECTORH          3		/* the "h" command (set signal high) */
#define VECTORX          4		/* the "x" command (set signal undefined) */
#define VECTORS          5		/* the "s" command (advance time) */
#define VECTORASSERT     6		/* the "assert" command (test signal value) */
#define VECTORCLOCK      7		/* the "clock" command to declare stimuli */
#define VECTORC          8		/* the "c" command to run a clock cycle */
#define VECTORVECTOR     9		/* the "vector" command to group signals */
#define VECTORSTEPSIZE  10		/* the "stepsize" command to set time advance */
#define VECTORPRINT     11		/* the "print" command to display messages */
#define VECTORSET       12		/* the "set" command to change vector values */
#define VECTOREXCL      13		/* the "!" command to print gate info */
#define VECTORQUESTION  14		/* the "?" command to print source/drain info */

typedef struct Isimvector
{
	INTBIG   command;
	INTBIG   parcount;
	CHAR   **parameters;
	struct Isimvector *nextsimvector;
} SIMVECTOR;

static SIMVECTOR *irsim_firstvector = NOSIMVECTOR;
static SIMVECTOR *irsim_lastvector = NOSIMVECTOR;
static SIMVECTOR *irsim_vectorfree = NOSIMVECTOR;


typedef	int	Coord;

typedef struct
{
	TimeType    first;
	TimeType    last;
	TimeType    start;
	TimeType    steps;
	TimeType    end;
} Times;

typedef struct TraceEnt *Trptr;

typedef struct
{
	int      disp;		/* number of traces displayed */
	Trptr    first;		/* ptr. to last trace displayed */
	Trptr    last;		/* list of traces */
} Traces;


typedef struct			/* Cache for history pointer */
{
	hptr  wind;			/* window start */
	hptr  cursor;		/* cursor value */
} Cache;

typedef struct TraceEnt
{
	Trptr    next;		/* doubly linked list of traces */
	Trptr    prev;
	CHAR     *name;		/* name stripped of path */
	INTBIG   waveformdata;
	int      len;		/* length of name string */
	Coord    top, bot;	/* position of the trace */
	short    bdigit;	/* # of bits per digit for displaying */
	short    vector;	/* 1 if bit vector, 0 if node */
	union
	{
		nptr    nd;		/* what makes up this trace */
		bptr    vec;
	} n;
	Cache    cache[1];
} TraceEnt;


/* get next history pointer, skiping punted events */
#define	NEXTH(H, P)	for((H) = (P)->next; (H)->punt; (H) = (H)->next)

#define	DEF_STEPS      4	/* default simulation steps per screen */
#define USERLINEWIDTH 50

static int      irsim_numAdded;
static Traces   irsim_traces;
static Times    irsim_tims;
static TimeType irsim_lastStart;		/* last redisplay starting time */
static INTBIG   irsim_window_iter;		/* iterator over signal names */

static double  *irsim_tracetime;
static INTSML  *irsim_tracestate;
static INTBIG   irsim_tracetotal = 0;

static CHAR     irsim_level_prefix[200];
static INTBIG   irsim_instnetnumber;

/* working memory for "irsim_loadvectorfile()" */
static INTBIG   irsim_inputparametertotal = 0;
static CHAR   **irsim_inputparameters;

/* prototypes for local routines */
static void       irsim_FlushTraceCache();
static void       irsim_DrawTraces(TimeType t1, TimeType t2);
static void       irsim_UpdateTraceCache(int first_trace);
static void       irsim_addtrace(Trptr t);
static void       irsim_helpwindow(BOOLEAN waveform);
static void       irsim_playvectors(void);
static void       irsim_newvector(INTBIG command, INTBIG parcount, CHAR **params, double inserttime);
static void       irsim_freevector(SIMVECTOR *sv);
static BOOLEAN    irsim_topofsignals(CHAR**);
static CHAR      *irsim_nextsignals(void);
static int        irsim_sortnetascending(const void *e1, const void *e2);
static int        irsim_sortbusnames(const void *e1, const void *e2);
static SIMVECTOR *irsim_allocsimvector(void);
static CHAR      *irsim_commandname(INTBIG command);
static double     irsim_endtime(void);
static void       irsim_showcurrentlevel(NODEPROTO *cell);
#if SIMFSDBWRITE != 0
static void       irsim_writeFsdb();
static int        irsim_writeNodeFsdb(nptr nd, Ulong ndindx, FILE *arg);
#endif

/*
 * Routine to free all memory allocated in this module.
 */
void irsim_freeanalyzermemory(void)
{
	REGISTER SIMVECTOR *sv;
	Trptr t;
	REGISTER INTBIG i;

	for(i=0; i<irsim_inputparametertotal; i++)
		efree((CHAR *)irsim_inputparameters[i]);
	if (irsim_inputparametertotal > 0)
		efree((CHAR *)irsim_inputparameters);
	irsim_inputparametertotal = 0;
	if (irsim_tracetotal > 0)
	{
		efree((CHAR *)irsim_tracetime);
		efree((CHAR *)irsim_tracestate);
		irsim_tracetotal = 0;
	}
	while (irsim_firstvector != NOSIMVECTOR)
	{
		sv = irsim_firstvector;
		irsim_firstvector = sv->nextsimvector;
		irsim_freevector(sv);
	}
	while (irsim_vectorfree != NOSIMVECTOR)
	{
		sv = irsim_vectorfree;
		irsim_vectorfree = sv->nextsimvector;
		efree((CHAR *)sv);
	}
	while (irsim_traces.first != 0)
	{
		t = irsim_traces.first;
		irsim_traces.first = t->next;
		efree((CHAR *)t);
	}
}

/************************** LOW-LEVEL ANALYZER **************************/

void irsim_addtrace(Trptr t)
{
	if (irsim_traces.first == NULL)
	{
		t->next = t->prev = NULL;
		irsim_traces.first = irsim_traces.last = t;
	} else
	{
		t->next = NULL;
		t->prev = irsim_traces.last;
		irsim_traces.last->next = t;
		irsim_traces.last = t;
	}
	irsim_numAdded++;
}

int irsim_AddNode(nptr nd, int *flag)
{
	Trptr  t;
	Q_UNUSED( flag );

	while (nd->nflags & ALIAS)
		nd = nd->nlink;

	if (nd->nflags & MERGED)
	{
		ttyputerr(x_("can't watch node %s"), nd->nname);
		return(1);
	}
	t = (Trptr)emalloc(sizeof(TraceEnt), sim_tool->cluster);
	if (t == NULL)
	{
		ttyputerr(x_("Out of memory, can't display '%s' in simulation"),
			nd->nname);
		return(0);
	}
	t->name = nd->nname;
	t->len = estrlen(t->name);
	t->bdigit = 1;
	t->vector = FALSE;
	t->n.nd = nd;
	nd->nflags &= ~HIST_OFF;
	t->cache[0].wind = t->cache[0].cursor = &(nd->head);
	irsim_addtrace(t);
	return(1);
}

int irsim_AddVector(bptr vec, int *flag)
{
	Trptr  t;
	int    n;

	n = vec->nbits;
	t = (Trptr) emalloc(sizeof(TraceEnt) + sizeof(Cache) * (n - 1), sim_tool->cluster);
	if (t == NULL)
	{
		ttyputerr(x_("Out of memory, can't add %s to analyzer"), vec->name);
		return(0);
	}
	t->name = vec->name;
	t->len = estrlen(t->name);
	if (*flag != 0)
		t->bdigit = *flag;
	else
		t->bdigit = (n > 4) ? 4 : 1;
	t->vector = TRUE;
	t->n.vec = vec;
	for(n--; n >= 0; n--)
	{
		vec->nodes[n]->nflags &= ~HIST_OFF;	
		t->cache[n].wind = t->cache[n].cursor = &(vec->nodes[n]->head);
	}
	irsim_addtrace(t);
	return(1);
}

void irsim_DisplayTraces(int isMapped)
{
	irsim_numAdded = 0;

	if (!isMapped)				/* only the first time */
	{
		irsim_FlushTraceCache();
	} else
	{
		irsim_traces.disp = sim_window_getnumframes();
		irsim_DrawTraces(irsim_tims.start, irsim_tims.end);
	}
}

void irsim_StopAnalyzer(void)
{
}

void irsim_RestartAnalyzer(TimeType first_time, TimeType last_time, int same_hist)
{
	register Trptr  t;
	register int    n;

	for(t = irsim_traces.first; t != 0; t = t->next)
	{
		if (t->vector)
		{
			for(n = t->n.vec->nbits - 1; n >= 0; n--)
			{
				t->cache[n].wind = t->cache[n].cursor = 
					&(t->n.vec->nodes[n]->head);
			}
		}
		else
			t->cache[0].wind = t->cache[0].cursor = &(t->n.nd->head);
	}

	/* should set "last_time" to the width of the screen */
	irsim_InitTimes(first_time, irsim_tims.steps / (TimeType)DEF_STEPS, last_time);
	if (same_hist)
		irsim_UpdateTraceCache(0);
	else
		irsim_FlushTraceCache();
}

static void RemoveTrace(Trptr t)
{
	if (t == irsim_traces.first)
	{
		irsim_traces.first = t->next;
		if (t->next)
			t->next->prev = NULL;
		else
			irsim_traces.last = NULL;
	} else
	{
		t->prev->next = t->next;
		if (t->next)
			t->next->prev = t->prev;
		else
			irsim_traces.last = t->prev;
	}

	efree((CHAR *)t);
}

void irsim_ClearTraces(void)
{
	while (irsim_traces.first != 0)
		RemoveTrace(irsim_traces.first);

	irsim_DrawTraces(irsim_tims.start, irsim_tims.end);
}

static void UpdateWinRemove()
{
	irsim_DrawTraces(irsim_tims.start, irsim_tims.end);
}

void irsim_RemoveVector(bptr b)
{
	Trptr t, tmp;
	BOOLEAN i = FALSE;

	for(t = irsim_traces.first; t != NULL;)
	{
		if (t->vector && (t->n.vec == b))
		{
			tmp = t->next;
			RemoveTrace(t);
			t = tmp;
			i = TRUE;
		}
		else
			t = t->next;
	}
	if (i)
		UpdateWinRemove();
}

void irsim_RemoveNode(nptr n)
{
	Trptr t, tmp;
	BOOLEAN i = FALSE;

	for(t = irsim_traces.first; t != NULL;)
	{
		if (!t->vector && (t->n.nd == n))
		{
			tmp = t->next;
			RemoveTrace(t);
			t = tmp;
			i = TRUE;
		} else
			t = t->next;
	}
	if (i) UpdateWinRemove();
}

void irsim_RemoveAllDeleted(void)
{
	Trptr t, tmp;
	BOOLEAN i = FALSE;

	for(t = irsim_traces.first; t != NULL;)
	{
		if ((t->vector && (t->n.vec->traced & DELETED)) ||
			(! t->vector && (t->n.nd->nflags & DELETED)))
		{
			tmp = t->next;
			RemoveTrace(t);
			t = tmp;
			i = TRUE;
		} else
			t = t->next;
	}
	if (i) UpdateWinRemove();
}

void irsim_TerminateAnalyzer(void)
{
}

/*
 * Initialize the windows and various other metrics.
 */
int irsim_InitDisplay(void)
{
	return(TRUE);
}

/*
 * Initialize the display times so that when first called the last time is
 * shown on the screen.  Default width is DEF_STEPS (simulation) steps.
 */
void irsim_InitTimes(TimeType firstT, TimeType stepsize, TimeType lastT)
{
	double min, max;
	INTBIG tr;
	TimeType endtime;

	irsim_tims.first = firstT;
	irsim_tims.last = lastT;
	irsim_tims.steps = 4 * stepsize;

	if (irsim_tims.start <= irsim_tims.first)
	{
		if (lastT < irsim_tims.steps)
		{
			irsim_tims.start = irsim_tims.first;
			irsim_tims.end = irsim_tims.start + irsim_tims.steps;

			/* make it conform to the displayed range */
			sim_window_inittraceloop();
			tr = sim_window_nexttraceloop();
			if (tr != 0)
			{
				sim_window_gettimerange(tr, &min, &max);
				endtime = ns2d(max * 1e9);
				if (endtime > irsim_tims.end) irsim_tims.end = endtime;
			}
			max = irsim_endtime();
			endtime = ns2d(max * 1e9);
			if (endtime > irsim_tims.end) irsim_tims.end = endtime;
		} else
		{
			irsim_tims.end = lastT + 2 * stepsize;
			irsim_tims.start = irsim_tims.end - irsim_tims.steps;
			if (irsim_tims.start < irsim_tims.first)
			{
				stepsize = irsim_tims.first - irsim_tims.start;
				irsim_tims.start += stepsize;
				irsim_tims.end += stepsize;
			}
		}
	}
}

/*
 * Update the cache (begining of window and cursor) for traces that just
 * became visible (or were just added).
 */
void irsim_UpdateTraceCache(int first_trace)
{
	register Trptr     t;
	register hptr      h,p;
	register int       n, i;
	register TimeType  startT, cursT;

	startT = irsim_tims.start;
	cursT = irsim_tims.first;
	for(t = irsim_traces.first, n = 0; t != 0; n++, t = t->next)
	{
		if (n < first_trace)
			continue;

		if (t->vector)
		{
			for(i = t->n.vec->nbits - 1; i >= 0; i--)
			{
				hptr  nexth;

				p = t->cache[i].wind;
				h = t->cache[i].cursor;
				NEXTH(nexth, h);
				if ((TimeType)h->htime > cursT || (TimeType)nexth->htime <= cursT)
				{
					if ((TimeType)p->htime <= cursT)	/* whatever is closer */
						t->cache[i].cursor = p;
					else
						t->cache[i].cursor = (hptr)&(t->n.vec->nodes[i]->head);
				}
				if (startT <= (TimeType)p->htime)			/* go back */
					p = (hptr) &(t->n.vec->nodes[i]->head);

				NEXTH(h, p);
				while ((TimeType)h->htime < startT)
				{
					p = h;
					NEXTH(h, h);
				}
				t->cache[i].wind = p;

				p = t->cache[i].cursor;
				NEXTH(h, p);
				while ((TimeType)h->htime <= cursT)
				{
					p = h;
					NEXTH(h, h);
				}
				t->cache[i].cursor = p;
			}
		} else
		{
			hptr  nexth;

			p = t->cache[0].wind;
			h = t->cache[0].cursor;
			NEXTH(nexth, h);
			if ((TimeType)h->htime > cursT || (TimeType)nexth->htime <= cursT)
			{
				if ((TimeType)p->htime <= cursT)
					t->cache[0].cursor = p;
				else
					t->cache[0].cursor = (hptr) &(t->n.nd->head);
			}

			if (startT <= (TimeType)p->htime)
				p = (hptr) &(t->n.nd->head);

			NEXTH(h, p);
			while ((TimeType)h->htime < startT)
			{
				p = h;
				NEXTH(h, h);
			}
			t->cache[0].wind = p;

			p = t->cache[0].cursor;
			NEXTH(h, p);
			while ((TimeType)h->htime <= cursT)
			{
				p = h;
				NEXTH(h, h);
			}
			t->cache[0].cursor = p;
		}
	}
}

void irsim_FlushTraceCache()
{
	irsim_lastStart = irsim_max_time;
}

/*
 * Draw the traces horizontally from time1 to time2.
 */
void irsim_DrawTraces(TimeType t1, TimeType t2)
{
	double *newtime;
	INTSML *newstate;
	INTBIG newtotal, i, count, val, n, begin;
	TimeType curt, endt, nextt;
	register Trptr   t;
	register TimeType  startT;
	register hptr      h, p;
#ifdef DUMPH_HOOK
	FILE *dumph;
#endif

	if (irsim_tims.start != irsim_lastStart)		/* Update history cache */
	{
		startT = irsim_tims.start;
		begin = (startT < irsim_lastStart);
		for(t = irsim_traces.first, n = irsim_traces.disp; t != 0; n--, t = t->next)
		{
			if (t == 0) break;
			if (t->vector)
			{
				for(i = t->n.vec->nbits - 1; i >= 0; i--)
				{
					p = begin ? (hptr) &(t->n.vec->nodes[i]->head) : t->cache[i].wind;
					NEXTH(h, p);
					while ((TimeType)h->htime < startT)
					{
						p = h;
						NEXTH(h, h);
					}
					t->cache[i].wind = p;
				}
			} else
			{
				p = begin ? (hptr) &(t->n.nd->head) : t->cache[0].wind;
				NEXTH(h, p);
				while ((TimeType)h->htime < startT)
				{
					p = h;
					NEXTH(h, h);
				}
				t->cache[0].wind = p;
			}
		}
		irsim_lastStart = irsim_tims.start;
	}

#ifdef DUMPH_HOOK
	dumph = fopen("dumph", "w");
#endif
	for(t = irsim_traces.first; t != 0; t = t->next)
	{
		if (t->vector != TRUE || t->n.vec->nbits <= 1)
		{
			if (t1 >= irsim_tims.last) continue;
#ifdef DUMPH_HOOK
			fprintf(dumph, "\n");
			fprintf(dumph, "Read ???.sim\n");
			fprintf(dumph, "\ttime  val    r/f\n");
			fprintf(dumph, "\n");
			fprintf(dumph, "%s\n", sim_irsimnets[t->waveformdata].name);
#endif

			h = t->cache[0].wind;
			count = 0;
			curt = 0;
			endt = t2;
			while (curt < endt)
			{
				val = h->val;
				while ((TimeType)h->htime < endt && h->val == (Uint)val)
					NEXTH(h, h);

				/* advance time */
				if ((TimeType)h->htime > endt)
				{
					nextt = endt;
				} else
				{
					nextt = (TimeType)h->htime;
				}

#ifdef DUMPH_HOOK
				if (h->inp) {
				  fprintf(dumph, "\t %12.2f %ci %11.2f\n", d2ns(nextt),
					  h->val == LOW ? '0' : h->val == HIGH ? '1' : 'X', 
					  d2ns(h->t.r.rtime));
				} else {
				  fprintf(dumph, "\t %12.2f %c %11.2f\n", d2ns(nextt),
					  h->val == LOW ? '0' : h->val == HIGH ? '1' : 'X', 
					  d2ns(h->t.r.rtime));
				}
#endif

				/* make sure there is room in the array */
				if (count >= irsim_tracetotal)
				{
					newtotal = irsim_tracetotal * 2;
					if (newtotal <= count) newtotal = count + 50;
					newtime = (double *)emalloc(newtotal * (sizeof (double)), sim_tool->cluster);
					if (newtime == 0) return;
					newstate = (INTSML *)emalloc(newtotal * SIZEOFINTSML, sim_tool->cluster);
					if (newstate == 0) return;
					for(i=0; i<count; i++)
					{
						newtime[i] = irsim_tracetime[i];
						newstate[i] = irsim_tracestate[i];
					}
					if (irsim_tracetotal > 0)
					{
						efree((CHAR *)irsim_tracetime);
						efree((CHAR *)irsim_tracestate);
					}
					irsim_tracetime = newtime;
					irsim_tracestate = newstate;
					irsim_tracetotal = newtotal;
				}

				irsim_tracetime[count] = curt / 100000000000.0;
				switch(val)
				{
					case LOW:	/* 0 */
						irsim_tracestate[count] = (LOGIC_LOW << 8) | GATE_STRENGTH;
						break;
					case HIGH:	/* 3 */
						irsim_tracestate[count] = (LOGIC_HIGH << 8) | GATE_STRENGTH;
						break;
					case X:		/* 1 */
						irsim_tracestate[count] = (LOGIC_X << 8) | GATE_STRENGTH;
						break;
				}
				curt = nextt;
				count++;
			}
		}
		sim_window_loaddigtrace(sim_irsimnets[t->waveformdata].signal, count, irsim_tracetime,
			irsim_tracestate);
	}
#ifdef DUMPH_HOOK
	fclose(dumph);
#endif
	sim_window_redraw();
}

static void UpdateTraces(TimeType start, TimeType end)
{
	irsim_DrawTraces(start, end);
}

/*
 * Update the trace window so that endT is shown.  If the update fits in the
 * window, simply draw the missing parts.  Otherwise scroll the traces,
 * centered around endT.
 */
void irsim_UpdateWindow(TimeType endT)
{
	TimeType  lastT;
	Trptr  t;
	int i;

	lastT = irsim_tims.last;
	irsim_tims.last = endT;

	for(t = irsim_traces.first; t != 0; t = t->next)
	{
		for(i=0; i<sim_irsimnetnumber; i++)
		{
			if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
			if (namesame(t->name, sim_irsimnets[i].name) == 0) break;
		}
		if (i < sim_irsimnetnumber)
			t->waveformdata = i;
	}

	if (endT <= irsim_tims.end)
	{
		if (lastT >= irsim_tims.start)
			UpdateTraces(lastT, endT);
		else if (endT > irsim_tims.start)
			UpdateTraces(irsim_tims.start, endT);
	} else					/* endT > irsim_tims.end */
	{
		if (lastT < irsim_tims.end)
			UpdateTraces(lastT, irsim_tims.end);
	}
}

/************************** ELECTRIC INTERFACE **************************/

void irsim_freememory(void)
{
	irsim_freeconfigmemory();
	irsim_freesimmemory();
	irsim_freeanalyzermemory();
	irsim_freersimmemory();
	irsim_freensubrsmemory();
}

BOOLEAN irsim_startsimulation(NODEPROTO *cell)
{
	REGISTER INTBIG i, sigvalid;
	INTBIG oldsigcount;
	REGISTER VARIABLE *var;
	CHAR testfile[300], line[USERLINEWIDTH], *truename, *pt;
	static BOOLEAN first = TRUE;
	REGISTER BOOLEAN showcommand;
	NODEPROTO *np;
	FILE *f;
	time_t ltime;
	nptr n;
	REGISTER void *infstr;

	/* first write the deck */
	estrcpy(testfile, x_("Electric.XXXXXX"));
	f = xcreate(testfile, sim_filetypeirsim|FILETYPETEMPFILE, 0, &truename);
	if (f == 0)
	{
		ttyputerr(_("Cannot write temporary simulation file '%s'"), testfile);
		return(TRUE);
	}
	sim_irsimgeneratedeck(cell, f);
	xclose(f);

	irsim_freememory();

	/* now initialize the simulator */
	time(&ltime);
	if (first)
	{
		irsim_InitThevs();
		irsim_InitCmdPath();
	}
	irsim_init_hist();
	irsim_stack_txtors = FALSE;
	ttyputmsg(x_("IRSIM %s"), irsim_version);

	/* read the configuration file */
	infstr = initinfstr();
	var = getval((INTBIG)sim_tool, VTOOL, VSTRING, x_("SIM_irsim_parameter_file"));
	if (var == NOVARIABLE)
		formatinfstr(infstr, x_("%s%s"), el_libdir, DEFIRSIMPARAMFILE); else
			addstringtoinfstr(infstr, (CHAR *)var->addr);
	if (irsim_config(returninfstr(infstr)) != 0) return(TRUE);

	/* Read network (.sim file) */
	if (irsim_rd_network(truename)) return(TRUE);

	/* remove the temporary network file */
	if (eunlink(truename) < 0)
		ttyputerr(_("Warning: cannot delete netlist file %s"), truename);

	irsim_ConnectNetwork();	/* connect all txtors to corresponding nodes */
	if (first)
		irsim_init_commands();	/* set up command table */
	irsim_init_event();
	first = FALSE;

	/* sort the signal names */
	esort(sim_irsimnets, sim_irsimnetnumber, sizeof (IRSIMNETWORK), irsim_sortnetascending);

	/* count the number of top-level signal names */
	sigvalid = 0;
	for(i=0; i<sim_irsimnetnumber; i++)
	{
		if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
		for(pt = sim_irsimnets[i].name; *pt != 0; pt++)
			if (*pt == '/') break;
		if (*pt != 0) continue;
		sigvalid++;
	}
	irsim_level_prefix[0] = 0;

	/* show the waveform window */
	irsim_firstvector = irsim_lastvector = NOSIMVECTOR;
	oldsigcount = 0;
	i = sim_window_isactive(&np);
	if ((i&SIMWINDOWWAVEFORM) == 0)
	{
		if (i != 0 && np != cell)
		{
			/* stop simulation of cell "np" */
			sim_window_stopsimulation();
		}

		/* remember signals that were last used with this cell */
		sim_window_grabcachedsignalsoncell(cell);

		/* show the simulation window */
		if (sim_window_create(sigvalid, cell,
			((sim_window_state&SHOWWAVEFORM) != 0 ? irsim_charhandlerwave : 0),
				irsim_charhandlerschem, IRSIM)) return(TRUE);
		sim_window_state = (sim_window_state & ~SIMENGINECUR) | SIMENGINECURIRSIM;
		sim_window_killalltraces(TRUE);
		irsim_showcurrentlevel(cell);
	}
	sim_window_cleartracehighlight();
	sim_window_settimerange(0, 0.0, DEFIRSIMTIMERANGE);
	sim_window_setmaincursor(DEFIRSIMTIMERANGE/5.0*2.0);
	sim_window_setextensioncursor(DEFIRSIMTIMERANGE/5.0*3.0);

	/* tell the simulator to watch all signals */
	if (! irsim_analyzerON)
	{
		if (! irsim_InitDisplay()) return(TRUE);
		irsim_InitTimes(irsim_sim_time0, 50000, irsim_cur_delta);
	}
	for(i=0; i<sim_irsimnetnumber; i++)
	{
		if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
		n = irsim_find(sim_irsimnets[i].name);
		if (n != 0)
			irsim_AddNode(n, 0);
	}
	irsim_UpdateWindow(0);
	irsim_DisplayTraces(irsim_analyzerON);		/* pass 0 first time */
	irsim_analyzerON = TRUE;

	/* disable automatic display */
	estrcpy(line, x_("display -automatic"));
	showcommand = FALSE;
	var = getvalkey((INTBIG)sim_tool, VTOOL, VINTEGER, sim_irsim_statekey);
	if (var != NOVARIABLE && (var->addr&IRSIMSHOWCOMMANDS) != 0) showcommand = TRUE;
	irsim_issuecommand(line, showcommand);

	/* read signal values */
	irsim_UpdateWindow(irsim_cur_delta);

	return(FALSE);
}

void irsim_showcurrentlevel(NODEPROTO *cell)
{
	REGISTER INTBIG i, j, k, l, len, buscount, oldsigcount;
	REGISTER CHAR *pt, *opt, *start, save, *name, *subname;
	CHAR **oldsignames, **bussigname;
	REGISTER void *infstr;
	REGISTER VARIABLE *var;
	INTBIG *bussignals;
	INTSML inistate[1];
	double initime[1], lowtime, hightime;
	Q_UNUSED( cell );

	/* remember the previous time range */
	sim_window_getaveragetimerange(&lowtime, &hightime);

	/* show the waveform window */
	irsim_ClearTraces();
	sim_window_cleartracehighlight();
	sim_window_killalltraces(FALSE);
	for(i=0; i<sim_irsimnetnumber; i++)
	{
		sim_irsimnets[i].flags &= ~IRSIMNETSHOWN;
		sim_irsimnets[i].signal = 0;
	}
	infstr = initinfstr();
	if (irsim_level_prefix[0] == 0)
		addstringtoinfstr(infstr, _("Top level")); else
	{
		len = estrlen(irsim_level_prefix);
		irsim_level_prefix[len-1] = 0;
		formatinfstr(infstr, _("level=%s"), irsim_level_prefix);
		irsim_level_prefix[len-1] = '/';
	}
	sim_window_titleinfo(returninfstr(infstr));
	sim_window_supresstraceprefix(irsim_level_prefix);

	initime[0] = 0.0;   inistate[0] = (LOGIC_X << 8) | GATE_STRENGTH;

	/* get signals saved from last time */
	oldsigcount = sim_window_getcachedsignals(&oldsignames);

	/* if the cell was from a different point in the hierarchy, ignore saved signals */
	var = getvalkey((INTBIG)cell, VNODEPROTO, VSTRING, sim_window_hierpos_key);
	if (var != NOVARIABLE)
	{
		if (namesame((CHAR *)var->addr, irsim_level_prefix) != 0)
		{
			if (getvalkey((INTBIG)cell, VNODEPROTO, VSTRING|VISARRAY,
				sim_window_signalorder_key) != NOVARIABLE)
					delvalkey((INTBIG)cell, VNODEPROTO, sim_window_signalorder_key);
			oldsigcount = 0;
		}
	}

	/* remember this position in the hierarchy */
	setvalkey((INTBIG)cell, VNODEPROTO, sim_window_hierpos_key, (INTBIG)irsim_level_prefix, VSTRING);

	/* show the signals saved from last time */
	for(j=0; j<oldsigcount; j++)
	{
		/* see if the name is a bus */
		for(pt = oldsignames[j]; *pt != 0; pt++) if (*pt == '\t') break;
		if (*pt == '\t')
		{
			/* a bus */
			pt++;
			sim_initbussignals();
			for(;;)
			{
				for(start = pt; *pt != 0; pt++) if (*pt == '\t') break;
				save = *pt;
				*pt = 0;
				opt = start;
				if (*opt == '-') opt++;
				for( ; *opt != 0; opt++)
					if (!isdigit(*opt) || *opt == ':') break;
				if (*opt == ':') start = opt+1;
				for(i=0; i<sim_irsimnetnumber; i++)
				{
					if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
					if ((sim_irsimnets[i].flags&IRSIMNETSHOWN) != 0) continue;
					name = sim_irsimnets[i].name;
					if (namesame(name, start) != 0) continue;
					sim_irsimnets[i].signal = sim_window_newtrace(-1, name, 0);
					sim_window_loaddigtrace(sim_irsimnets[i].signal, 1, initime, inistate);
					sim_irsimnets[i].flags |= IRSIMNETSHOWN;
					sim_addbussignal(sim_irsimnets[i].signal);
					break;
				}
				*pt++ = save;
				if (save == 0) break;
			}

			/* create the bus in the waveform window */
			infstr = initinfstr();
			for(pt = oldsignames[j]; *pt != 0; pt++)
			{
				if (*pt == '\t') break;
				addtoinfstr(infstr, *pt);
			}
			pt = returninfstr(infstr);
			buscount = sim_getbussignals(&bussignals);
			(void)sim_window_makebus(buscount, bussignals, pt);

			/* make the bus in IRSIM */
			bussigname = (CHAR **)emalloc((buscount+1) * (sizeof (CHAR *)), sim_tool->cluster);
			if (bussigname == 0) return;
			for(i=0; i<buscount; i++) bussigname[i+1] = sim_window_gettracename(bussignals[i]);
			bussigname[0] = pt;
			irsim_newvector(VECTORVECTOR, buscount+1, bussigname, 0.0);
			efree((CHAR *)bussigname);
		} else
		{
			/* a single signal */
			pt = oldsignames[j];
			if (*pt == '-') pt++;
			for( ; *pt != 0; pt++)
				if (!isdigit(*pt) || *pt == ':') break;
			if (*pt == ':') pt++; else pt = oldsignames[j];
			for(i=0; i<sim_irsimnetnumber; i++)
			{
				if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
				if ((sim_irsimnets[i].flags&IRSIMNETSHOWN) != 0) continue;
				name = sim_irsimnets[i].name;
				if (namesame(name, pt) != 0) continue;

				sim_irsimnets[i].signal = sim_window_newtrace(-1, name, (INTBIG)0);
				sim_window_loaddigtrace(sim_irsimnets[i].signal, 1, initime, inistate);
				sim_irsimnets[i].flags |= IRSIMNETSHOWN;
				break;
			}
		}
	}

	/* show default signals if none cached */
	if (oldsigcount == 0)
	{
		len = estrlen(irsim_level_prefix);
		for(i=0; i<sim_irsimnetnumber; i++)
		{
			if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
			if ((sim_irsimnets[i].flags&IRSIMNETSHOWN) != 0) continue;
			name = sim_irsimnets[i].name;
			if (len != 0)
			{
				/* down a level: the prefix must match */
				if (namesamen(name, irsim_level_prefix, len) != 0) continue;
			}

			/* there must be no further hierarchy in the name */
			for(l=len; name[l] != 0; l++) if (name[l] == '/') break;
			if (name[l] != 0) continue;

			/* see if the signal is arrayed */
			for(l=0; name[l] != 0; l++) if (name[l] == '[') break;
			if (name[l] == '[')
			{
				/* found an arrayed signal: gather it into a bus */
				sim_initbussignals();
				for(j=0; j<sim_irsimnetnumber; j++)
				{
					if ((sim_irsimnets[j].flags&IRSIMNETVALID) == 0) continue;
					if ((sim_irsimnets[j].flags&IRSIMNETSHOWN) != 0) continue;
					subname = sim_irsimnets[j].name;
					if (namesamen(subname, name, l) != 0) continue;
					if (subname[l] != 0 && subname[l] != '[') continue;
					sim_addbussignal(j);
				}
				buscount = sim_getbussignals(&bussignals);
				bussigname = (CHAR **)emalloc((buscount+1) * (sizeof (CHAR *)), sim_tool->cluster);
				if (bussigname == 0) return;
				esort(bussignals, buscount, SIZEOFINTBIG, irsim_sortbusnames);
				for(j=0; j<buscount; j++)
				{
					k = bussignals[j];
					subname = sim_irsimnets[k].name;
					bussignals[j] = sim_window_newtrace(-1, subname, 0);
					bussigname[j+1] = subname;
					sim_window_loaddigtrace(bussignals[j], 1, initime, inistate);
					sim_irsimnets[k].signal = bussignals[j];
					sim_irsimnets[k].flags |= IRSIMNETSHOWN;
				}

				/* create the bus */
				if (buscount > 1)
				{
					name[l] = 0;
					infstr = initinfstr();
					addstringtoinfstr(infstr, name);
					bussigname[0] = returninfstr(infstr);
					name[l] = '[';
					(void)sim_window_makebus(buscount, bussignals, bussigname[0]);
					irsim_newvector(VECTORVECTOR, buscount+1, bussigname, 0.0);
				}
				efree((CHAR *)bussigname);
			} else
			{
				sim_irsimnets[i].signal = sim_window_newtrace(-1, name, 0);
				sim_irsimnets[i].flags |= IRSIMNETSHOWN;
				sim_window_loaddigtrace(sim_irsimnets[i].signal, 1, initime, inistate);
			}
		}
	}

	/* now set the time range to the previous state */
	sim_window_settimerange(0, lowtime, hightime);

	/* sim_window_redraw(); */
}

/*
 * Helper routine for "esort" that makes bus signals be numerically ordered
 */
int irsim_sortbusnames(const void *e1, const void *e2)
{
	REGISTER CHAR *n1, *n2;
	REGISTER INTBIG i1, i2;

	i1 = *((INTBIG *)e1);
	i2 = *((INTBIG *)e2);
	n1 = sim_irsimnets[i1].name;
	n2 = sim_irsimnets[i2].name;
	if ((net_options&NETDEFBUSBASEDESC) != 0)
		return(namesamenumeric(n2, n1));
	return(namesamenumeric(n1, n2));
}

/*
 * Helper routine for "esort" that makes IRSIM network objects go in ascending name order.
 */
int irsim_sortnetascending(const void *e1, const void *e2)
{
	REGISTER IRSIMNETWORK *in1, *in2;

	in1 = (IRSIMNETWORK *)e1;
	in2 = (IRSIMNETWORK *)e2;
	return(namesame(in1->name, in2->name));
}

/*
 * The character handler for the schematic/layout window of ALS simulation
 */
BOOLEAN irsim_charhandlerschem(WINDOWPART *w, INTSML chr, INTBIG special)
{
	INTBIG i, *highsiglist, highsigcount, veccmd, *tracelist;
	REGISTER NETWORK *net, **netlist;
	REGISTER BOOLEAN showcommand;
	REGISTER VARIABLE *var;
	CHAR *pt, *highsigname, *par[1];
	REGISTER SIMVECTOR *sv, *lastsv, *nextsv;
	REGISTER void *infstr;

	ttynewcommand();

	/* special characters are not handled here */
	if (special != 0)
		return(us_charhandler(w, chr, special));

	switch (chr)
	{
		case '?':
			irsim_helpwindow(FALSE);
			return(FALSE);
		case '0':
		case 'l':
			veccmd = VECTORL;
			break;
		case '1':
		case 'h':
			veccmd = VECTORH;
			break;
		case 'x':
			veccmd = VECTORX;
			break;
		case 'i':
		case ' ':
		case 'e':
		case 'a':
			break;
		default:
			return(us_charhandler(w, chr, special));
	}

	/* find the net in the waveform window */
	netlist = net_gethighlightednets(FALSE);
	net = netlist[0];
	if (net == NONETWORK) return(FALSE);
	if (net->namecount > 0) pt = networkname(net, 0); else
		pt = describenetwork(net);
	(void)allocstring(&highsigname, pt, sim_tool->cluster);
	if (chr == 'a')
	{
		irsim_adddisplayedsignal(highsigname);
		efree((CHAR *)highsigname);
		return(FALSE);
	}
	highsiglist = sim_window_findtrace(highsigname, &highsigcount);
	if (highsigcount == 0)
	{
		ttyputerr(_("Cannot find network %s in simulation"), highsigname);
		efree((CHAR *)highsigname);
		return(FALSE);
	}
	sim_window_cleartracehighlight();
	for(i=0; i<highsigcount; i++)
		sim_window_addhighlighttrace(highsiglist[i]);

	if (chr == 'e')
	{
		lastsv = NOSIMVECTOR;
		for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = nextsv)
		{
			nextsv = sv->nextsimvector;
			if (sv->command == VECTORL || sv->command == VECTORH || sv->command == VECTORX ||
				sv->command == VECTORASSERT || sv->command == VECTORSET)
			{
				if (namesame(sv->parameters[0], highsigname) == 0)
				{
					if (lastsv == NOSIMVECTOR)
						irsim_firstvector = sv->nextsimvector; else
							lastsv->nextsimvector = sv->nextsimvector;
					irsim_freevector(sv);
					continue;
				}
			}
			lastsv = sv;
		}
		irsim_lastvector = lastsv;
		irsim_playvectors();
		efree((CHAR *)highsigname);
		return(FALSE);
	}
	if (chr == ' ')		/* issue IRSIM command */
	{
		pt = ttygetlinemessages(x_("> "));
		irsim_issuecommand(pt, FALSE);
		efree((CHAR *)highsigname);
		return(FALSE);
	}
	if (chr == 'i')
	{
		showcommand = FALSE;
		var = getvalkey((INTBIG)sim_tool, VTOOL, VINTEGER, sim_irsim_statekey);
		if (var != NOVARIABLE && (var->addr&IRSIMSHOWCOMMANDS) != 0) showcommand = TRUE;

		infstr = initinfstr();
		formatinfstr(infstr, x_("! %s"), highsigname);
		irsim_issuecommand(returninfstr(infstr), showcommand);
		infstr = initinfstr();
		formatinfstr(infstr, x_("? %s"), highsigname);
		irsim_issuecommand(returninfstr(infstr), showcommand);
		efree((CHAR *)highsigname);
		return(FALSE);
	}

	if (highsigcount > 1)
	{
		ttyputmsg(_("Cannot set level on multiple signals"));
		return(FALSE);
	}
	tracelist = sim_window_getbustraces(highsiglist[0]);
	if (tracelist != 0 && tracelist[0] != 0)
	{
		ttyputmsg(_("Cannot set level on a bus, use 'v' to set a value"));
		return(FALSE);
	}
	par[0] = highsigname;
	irsim_newvector(veccmd, 1, par, sim_window_getmaincursor());
	irsim_playvectors();
	efree((CHAR *)highsigname);
	return(FALSE);
}

BOOLEAN irsim_topofsignals(CHAR **c)
{
	Q_UNUSED( c );

	irsim_window_iter = 0;
	return(TRUE);
}

CHAR *irsim_nextsignals(void)
{
	CHAR *nextname;

	for(; ; irsim_window_iter++)
	{
		if (irsim_window_iter >= sim_irsimnetnumber) return(0);
		if ((sim_irsimnets[irsim_window_iter].flags&IRSIMNETVALID) == 0) continue;
		if ((sim_irsimnets[irsim_window_iter].flags&IRSIMNETSHOWN) == 0) break;
	}
	nextname = sim_irsimnets[irsim_window_iter].name;
	irsim_window_iter++;
	return(nextname);
}

static COMCOMP sim_irsimpicktrace = {NOKEYWORD, irsim_topofsignals,
	irsim_nextsignals, NOPARAMS, 0, x_(" \t"), N_("pick a signal to display"), x_("")};

/*
 * The character handler for the waveform window of ALS simulation
 */
BOOLEAN irsim_charhandlerwave(WINDOWPART *w, INTSML chr, INTBIG special)
{
	NODEPROTO *np;
	REGISTER INTBIG *highsigs, highsig, i, j, thispos, *bussigs,
		trl, nexttr, prevtr, pos, bitoffset, veccmd, numbits, buscount;
	INTBIG state, *tracelist, *theBits;
	REGISTER BOOLEAN showcommand, foundsig;
	REGISTER VARIABLE *var;
	CHAR *msg, *par[30], setvalue[MAXSIMWINDOWBUSWIDTH+1];
	REGISTER SIMVECTOR *sv, *lastsv, *nextsv;
	REGISTER void *infstr;

	ttynewcommand();

	/* special characters are not handled here */
	if (special != 0)
		return(us_charhandler(w, chr, special));

	/* can always preserve snapshot */
	if (chr == 'p')
	{
		sim_window_savegraph();
		return(FALSE);
	}

	/* can always do help */
	if (chr == '?')
	{
		irsim_helpwindow(TRUE);
		return(FALSE);
	}

	if (chr == ' ')		/* issue IRSIM command */
	{
		msg = ttygetlinemessages(x_("> "));
		irsim_issuecommand(msg, FALSE);
		return(FALSE);
	}

	/* if not simulating, don't handle any simulation commands */
	if (sim_window_isactive(&np) == 0)
		return(us_charhandler(w, chr, special));
	veccmd = 0;
	numbits = 0;
	switch (chr)
	{
		/* convert busses */
		case 'b':
			if (sim_window_buscommand()) return(FALSE);
			highsigs = sim_window_gethighlighttraces();
#if 0	/* should enter a "vector" command here */
			{
				CHAR *bussigname[MAXSIMWINDOWBUSWIDTH+1];
				bussigname[0] = pt;
				irsim_newvector(VECTORVECTOR, buscount+1, bussigname, 0.0);
			}
#endif
			irsim_UpdateWindow(irsim_cur_delta);
			return(FALSE);

		/* add trace */
		case 'a':
			i = ttygetparam(_("Signal to add"), &sim_irsimpicktrace, 3, par);
			if (i == 0) return(FALSE);
			irsim_adddisplayedsignal(par[0]);
			return(FALSE);

		/* different flavors of low values */
		case '0':
		case 'l':
			state = LOGIC_LOW;
			veccmd = VECTORL;
			break;

		/* different flavors of high values */
		case '1':
		case 'h':
			state = LOGIC_HIGH;
			veccmd = VECTORH;
			break;

		/* different flavors of undefined values */
		case 'x':
			state = LOGIC_X;
			veccmd = VECTORX;
			break;

		/* different flavors of wide numeric values */
		case 'v':
			numbits = sim_window_getwidevalue(&theBits);
			if (numbits < 0) return(FALSE);
			break;

		/* signal clock setting, info, erasing, removing (all handled later) */
		case 'e':
		case 'r':
		case DELETEKEY:
		case 'i':
		case ' ':
			break;

		default:
			return(us_charhandler(w, chr, special));
	}

	/* the following commands demand a current trace...get it */
	highsigs = sim_window_gethighlighttraces();
	if (highsigs[0] == 0)
	{
		ttyputerr(_("Select a signal name first"));
		return(FALSE);
	}
	if (chr == 'r' || chr == DELETEKEY)		/* remove trace(s) */
	{
		sim_window_cleartracehighlight();

		/* delete them */
		nexttr = prevtr = 0;
		for(j=0; highsigs[j] != 0; j++)
		{
			highsig = highsigs[j];
			thispos = sim_window_gettraceframe(highsig);
			bussigs = sim_window_getbustraces(highsig);
			for(buscount=0; bussigs[buscount] != 0; buscount++) ;
			sim_window_inittraceloop();
			nexttr = prevtr = 0;
			for(;;)
			{
				trl = sim_window_nexttraceloop();
				if (trl == 0) break;
				pos = sim_window_gettraceframe(trl);
				if (pos > thispos)
				{
					if (pos-1 == thispos) nexttr = trl;
					pos = pos - 1 + buscount;
					sim_window_settraceframe(trl, pos);
				} else if (pos == thispos-1) prevtr = trl;
			}
			if (buscount > 0)
			{
				for(i=0; i<buscount; i++)
					sim_window_settraceframe(bussigs[i], thispos+i);
			}

			/* remove from the simulator's list */
			for(i=0; i<sim_irsimnetnumber; i++)
			{
				if (sim_irsimnets[i].signal == highsig)
				{
					sim_irsimnets[i].signal = 0;
					break;
				}
			}

			/* kill trace */
			sim_window_killtrace(highsig);
		}

		/* redraw */
		if (nexttr != 0)
		{
			sim_window_addhighlighttrace(nexttr);
		} else if (prevtr != 0)
		{
			sim_window_addhighlighttrace(prevtr);
		}
		sim_window_redraw();
		return(FALSE);
	}

	if (highsigs[1] != 0)
	{
		ttyputerr(_("Select just one signal name first"));
		return(FALSE);
	}
	highsig = highsigs[0];

	if (chr == 'i')		/* print signal info */
	{
		showcommand = FALSE;
		var = getvalkey((INTBIG)sim_tool, VTOOL, VINTEGER, sim_irsim_statekey);
		if (var != NOVARIABLE && (var->addr&IRSIMSHOWCOMMANDS) != 0) showcommand = TRUE;

		infstr = initinfstr();
		formatinfstr(infstr, x_("! %s"), sim_window_gettracename(highsig));
		irsim_issuecommand(returninfstr(infstr), showcommand);
		infstr = initinfstr();
		formatinfstr(infstr, x_("? %s"), sim_window_gettracename(highsig));
		irsim_issuecommand(returninfstr(infstr), showcommand);
		return(FALSE);
	}
	if (chr == 'e')		/* clear signal vectors */
	{
		tracelist = sim_window_getbustraces(highsig);
		lastsv = NOSIMVECTOR;
		for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = nextsv)
		{
			nextsv = sv->nextsimvector;
			if (sv->command == VECTORL || sv->command == VECTORH || sv->command == VECTORX ||
				sv->command == VECTORASSERT || sv->command == VECTORSET)
			{
				foundsig = FALSE;
				if (namesame(sv->parameters[0], sim_window_gettracename(highsig)) == 0)
					foundsig = TRUE; else
				{
					if (tracelist != 0 && tracelist[0] != 0)
					{
						for(i=0; tracelist[i] != 0; i++)
						{
							if (namesame(sv->parameters[0], sim_window_gettracename(tracelist[i])) == 0)
							{
								foundsig = TRUE;
								break;
							}
						}
					}
				}
				if (foundsig)
				{
					if (lastsv == NOSIMVECTOR)
						irsim_firstvector = sv->nextsimvector; else
							lastsv->nextsimvector = sv->nextsimvector;
					irsim_freevector(sv);
					continue;
				}
			}
			lastsv = sv;
		}
		irsim_lastvector = lastsv;
		irsim_playvectors();
		return(FALSE);
	}

	/* handle setting of values on signals */
	if (chr == 'v')
	{
		tracelist = sim_window_getbustraces(highsig);
		if (tracelist == 0 || tracelist[0] == 0)
		{
			ttyputerr(_("Select a bus signal before setting numeric values on it"));
			return(FALSE);
		}
		for(i=0; tracelist[i] != 0; i++) ;
		bitoffset = numbits - i;
		for(i=0; tracelist[i] != 0; i++)
		{
			if (i+bitoffset < 0 || theBits[i+bitoffset] == 0) setvalue[i] = '0'; else
				setvalue[i] = '1';
		}
		setvalue[i] = 0;
		par[0] = sim_window_gettracename(highsig);
		par[1] = setvalue;
		irsim_newvector(VECTORSET, 2, par, sim_window_getmaincursor());
	} else
	{
		tracelist = sim_window_getbustraces(highsig);
		if (tracelist != 0 && tracelist[0] != 0)
		{
			ttyputmsg(_("Cannot set level on a bus, use 'v' to set a value"));
			return(FALSE);
		}
		par[0] = sim_window_gettracename(highsig);
		irsim_newvector(veccmd, 1, par, sim_window_getmaincursor());
	}
	irsim_playvectors();
	return(FALSE);
}

void irsim_helpwindow(BOOLEAN waveform)
{
	REGISTER INTBIG active;
	NODEPROTO *np;

	active = sim_window_isactive(&np);
	if (active == 0)
	{
		ttyputmsg(_("There is no current simulation"));
		return;
	}

	if (waveform)
		ttyputmsg(_("These keys may be typed in the IRSIM Waveform window:")); else
			ttyputmsg(_("These keys may be typed in the IRSIM Schematic window:"));
	ttyputinstruction(x_("l,0"),  5, _("Set signal low"));
	ttyputinstruction(x_("h,1"),  5, _("Set signal high"));
	ttyputinstruction(x_(" x"),   5, _("Set signal undefined"));
	if (waveform)
	{
		ttyputinstruction(x_(" v"),  5, _("Set bus to wide value"));
	}
	ttyputinstruction(x_(" e"),   5, _("Delete all vectors on signal"));
	ttyputinstruction(x_(" i"),   5, _("Print information about signal"));
	ttyputinstruction(x_("<sp>"), 5, _("Issue IRSIM command"));
	ttyputinstruction(x_(" a"),   5, _("Add signal to simulation window"));
	if (waveform)
	{
		ttyputinstruction(x_(" r"), 5, _("Remove signal from the window"));
		ttyputinstruction(x_(" b"), 5, _("Combine selected signals into a bus"));
		ttyputinstruction(x_(" p"), 5, _("Preserve snapshot of simulation window in database"));
	}
}

/************************** SIMULATION VECTORS **************************/

SIMVECTOR *irsim_allocsimvector(void)
{
	REGISTER SIMVECTOR *sv;

	if (irsim_vectorfree != NOSIMVECTOR)
	{
		sv = irsim_vectorfree;
		irsim_vectorfree = sv->nextsimvector;
	} else
	{
		sv = (SIMVECTOR *)emalloc(sizeof (SIMVECTOR), sim_tool->cluster);
		if (sv == 0) return(NOSIMVECTOR);
	}
	return(sv);
}

/*
 * Routine to create a new simulation vector at time "time", on signal "sig",
 * with state "state".  The vector is inserted into the play list in the proper
 * order.
 */
void irsim_newvector(INTBIG command, INTBIG parcount, CHAR **params, double inserttime)
{
	REGISTER SIMVECTOR *sv, *newsv, *lastsv, *aftersv;
	REGISTER INTBIG i, clockphases;
	CHAR line[200];
	double curtime, defaultstepsize, stepsize, thisstep;

	newsv = irsim_allocsimvector();
	if (newsv == NOSIMVECTOR) return;
	newsv->command = command;
	newsv->parcount = parcount;
	if (parcount > 0)
	{
		newsv->parameters = (CHAR **)emalloc(parcount * (sizeof (CHAR *)), sim_tool->cluster);
		if (newsv->parameters == 0) return;
		for(i=0; i<parcount; i++)
			(void)allocstring(&newsv->parameters[i], params[i], sim_tool->cluster);
	}

	/* insert the vector */
	lastsv = NOSIMVECTOR;
	if (inserttime < 0.0)
	{
		lastsv = irsim_lastvector;
	} else
	{
		defaultstepsize = 10.0 / 1000000000.0;
		curtime = 0.0;
		clockphases = 0;
		for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = sv->nextsimvector)
		{
			switch (sv->command)
			{
				case VECTORCLOCK:
					clockphases = sv->parcount - 1;
					break;
				case VECTORS:
					if (sv->parcount <= 0) stepsize = defaultstepsize; else
						stepsize = eatof(sv->parameters[0]) / 1000000000.0;
					if (doubleslessthan(inserttime, curtime+stepsize))
					{
						/* splitting step at "inserttime" */
						esnprintf(line, 200, x_("%g"), (inserttime-curtime) * 1000000000.0);
						if (sv->parcount == 0)
						{
							sv->parameters = (CHAR **)emalloc((sizeof (CHAR *)),
								sim_tool->cluster);
							(void)allocstring(&sv->parameters[0], line, sim_tool->cluster);
						} else
						{
							(void)reallocstring(&sv->parameters[0], line, sim_tool->cluster);
						}

						/* create second step to advance after this signal */
						aftersv = irsim_allocsimvector();
						if (aftersv == NOSIMVECTOR) return;
						aftersv->command = VECTORS;
						aftersv->parcount = 1;
						aftersv->parameters = (CHAR **)emalloc((sizeof (CHAR *)),
							sim_tool->cluster);
						if (aftersv->parameters == 0) return;
						esnprintf(line, 200, x_("%g"), (curtime+stepsize-inserttime) * 1000000000.0);
						(void)allocstring(&aftersv->parameters[0], line, sim_tool->cluster);
						aftersv->nextsimvector = sv->nextsimvector;
						sv->nextsimvector = aftersv;
					}
					curtime += stepsize;
					break;
				case VECTORC:
					if (sv->parcount <= 0) stepsize = defaultstepsize; else
						stepsize = eatof(sv->parameters[0]) / 1000000000.0;
					curtime += stepsize * clockphases;
					break;
				case VECTORSTEPSIZE:
					if (sv->parcount > 0)
						defaultstepsize = eatof(sv->parameters[0]) / 1000000000.0;
					break;
			}
			lastsv = sv;
			if (!doubleslessthan(curtime, inserttime)) break;
		}
		if (doubleslessthan(curtime, inserttime))
		{
			/* create step to advance to the insertion time */
			thisstep = (inserttime-curtime) * 1000000000.0;
			if (thisstep > 0.005)
			{
				aftersv = irsim_allocsimvector();
				if (aftersv == NOSIMVECTOR) return;
				aftersv->command = VECTORS;
				aftersv->parcount = 1;
				aftersv->parameters = (CHAR **)emalloc((sizeof (CHAR *)),
					sim_tool->cluster);
				if (aftersv->parameters == 0) return;
				esnprintf(line, 200, x_("%g"), thisstep);
				(void)allocstring(&aftersv->parameters[0], line, sim_tool->cluster);
				if (lastsv == NOSIMVECTOR)
				{
					aftersv->nextsimvector = irsim_firstvector;
					irsim_firstvector = aftersv;
				} else
				{
					aftersv->nextsimvector = lastsv->nextsimvector;
					lastsv->nextsimvector = aftersv;
				}
				lastsv = aftersv;
			}
		}
	}
	if (lastsv == NOSIMVECTOR)
	{
		newsv->nextsimvector = irsim_firstvector;
		irsim_firstvector = newsv;
	} else
	{
		newsv->nextsimvector = lastsv->nextsimvector;
		lastsv->nextsimvector = newsv;
	}
	if (newsv->nextsimvector == NOSIMVECTOR)
	{
		irsim_lastvector = newsv;
	}
}

/*
 * Routine to examine the test vectors and determine the ending simulation time.
 */
double irsim_endtime(void)
{
	REGISTER SIMVECTOR *sv;
	REGISTER INTBIG clockphases;
	double curtime, defaultstepsize, stepsize;

	defaultstepsize = 10.0 / 1000000000.0;
	curtime = 0.0;
	clockphases = 0;
	for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = sv->nextsimvector)
	{
		switch (sv->command)
		{
			case VECTORCLOCK:
				clockphases = sv->parcount - 1;
				break;
			case VECTORS:
				if (sv->parcount <= 0) stepsize = defaultstepsize; else
					stepsize = eatof(sv->parameters[0]) / 1000000000.0;
				curtime += stepsize;
				break;
			case VECTORC:
				if (sv->parcount <= 0) stepsize = defaultstepsize; else
					stepsize = eatof(sv->parameters[0]) / 1000000000.0;
				curtime += stepsize * clockphases;
				break;
			case VECTORSTEPSIZE:
				if (sv->parcount > 0)
					defaultstepsize = eatof(sv->parameters[0]) / 1000000000.0;
				break;
		}
	}
	return(curtime);
}

/*
 * Routine to free simulation vector "sv"
 */
void irsim_freevector(SIMVECTOR *sv)
{
	REGISTER INTBIG i;

	for(i=0; i<sv->parcount; i++) efree((CHAR *)sv->parameters[i]);
	if (sv->parcount > 0) efree((CHAR *)sv->parameters);
	sv->nextsimvector = irsim_vectorfree;
	irsim_vectorfree = sv;
}

/*
 * Routine to play the simulation vectors into the simulator.
 */
void irsim_playvectors(void)
{
	REGISTER SIMVECTOR *sv;
	REGISTER INTBIG i;
	double curtime;
	REGISTER BOOLEAN showcommand;
	REGISTER VARIABLE *var;
	REGISTER void *infstr;

	/* get IRSIM state */
	showcommand = FALSE;
	var = getvalkey((INTBIG)sim_tool, VTOOL, VINTEGER, sim_irsim_statekey);
	if (var != NOVARIABLE && (var->addr&IRSIMSHOWCOMMANDS) != 0) showcommand = TRUE;

	irsim_issuecommand(x_("back 0"), showcommand);
	irsim_issuecommand(x_("flush"), showcommand);

	curtime = 0;
	irsim_analyzerON = FALSE;
	for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = sv->nextsimvector)
	{
		if (sv->command == VECTORCOMMENT) continue;
		infstr = initinfstr();
		addstringtoinfstr(infstr, irsim_commandname(sv->command));
		for(i=0; i<sv->parcount; i++)
			formatinfstr(infstr, x_(" %s"), sv->parameters[i]);
		irsim_issuecommand(returninfstr(infstr), showcommand);
	}
	irsim_issuecommand(x_("s"), showcommand);
	irsim_analyzerON = TRUE;
	irsim_UpdateWindow(irsim_cur_delta);

	/* update main cursor location if requested */
	if ((sim_window_state&ADVANCETIME) != 0)
		sim_window_setmaincursor(curtime + 10.0/1000000000.0);
#if SIMFSDBWRITE != 0
	irsim_writeFsdb();
#endif
}

CHAR *irsim_commandname(INTBIG command)
{
	switch (command)
	{
		case VECTORL:        return(x_("l"));
		case VECTORH:        return(x_("h"));
		case VECTORX:        return(x_("x"));
		case VECTORS:        return(x_("s"));
		case VECTORASSERT:   return(x_("assert"));
		case VECTORCLOCK:    return(x_("clock"));
		case VECTORC:        return(x_("c"));
		case VECTORVECTOR:   return(x_("vector"));
		case VECTORSTEPSIZE: return(x_("stepsize"));
		case VECTORPRINT:    return(x_("print"));
		case VECTORSET:      return(x_("set"));
		case VECTOREXCL:     return(x_("!"));
		case VECTORQUESTION: return(x_("?"));
	}
	return(x_(""));
}

/*
 * Routine to clear all simulation vectors.
 */
void irsim_clearallvectors(void)
{
	REGISTER SIMVECTOR *sv;

	while (irsim_firstvector != NOSIMVECTOR)
	{
		sv = irsim_firstvector;
		irsim_firstvector = sv->nextsimvector;
		irsim_freevector(sv);
	}
	irsim_lastvector = NOSIMVECTOR;
	irsim_playvectors();
}

/*
 * Routine to read simulation vectors from file "filename" (if zero, prompt for file).
 */
void irsim_loadvectorfile(CHAR *filename)
{
	CHAR *truename, buf[200], *pt, *keyword, *par[1], **newparamlist;
	FILE *vin;
	REGISTER SIMVECTOR *sv;
	double min, max;
	REGISTER INTBIG lineno, command, count, newtotal, i, j, *bussignals, *highsiglist;
	INTBIG highsigcount;

	if (filename == 0)
	{
		filename = fileselect(_("IRSIM vector file"), sim_filetypeirsimcmd, x_(""));
		if (filename == 0) return;
	}
	vin = xopen(filename, sim_filetypeirsimcmd, x_(""), &truename);
	if (! vin)
	{
		ttyputerr(_("ERROR: Can't open %s"), truename);
		return;
	}

	/* remove all vectors */
	while (irsim_firstvector != NOSIMVECTOR)
	{
		sv = irsim_firstvector;
		irsim_firstvector = sv->nextsimvector;
		irsim_freevector(sv);
	}
	irsim_lastvector = NOSIMVECTOR;
	ttyputmsg(_("Reading %s"), truename);
	lineno = 0;
	for(;;)
	{
		if (xfgets(buf, 200, vin)) break;
		lineno++;

		/* ignore comments */
		if (buf[0] == '|')
		{
			par[0] = buf;
			irsim_newvector(VECTORCOMMENT, 1, par, -1.0);
			continue;
		}

		/* get the first keyword */
		pt = buf;
		keyword = getkeyword(&pt, x_(" "));
		if (keyword == NOSTRING || *keyword == 0)
		{
			par[0] = buf;
			irsim_newvector(VECTORCOMMENT, 1, par, -1.0);
			continue;
		}

		/* ignore the "w" command */
		if (namesame(keyword, x_("w")) == 0)
		{
			par[0] = buf;
			irsim_newvector(VECTORCOMMENT, 1, par, -1.0);
			continue;
		}

		/* save the "print" command */
		if (namesame(keyword, x_("print")) == 0)
		{
			while (*pt == ' ') pt++;
			par[0] = pt;
			irsim_newvector(VECTORPRINT, 1, par, -1.0);
			continue;
		}

		/* allow interruption */
		if (stopping(STOPREASONCOMFILE)) break;

		/* handle level setting on signals */
		if (namesame(keyword, x_("l")) == 0 || namesame(keyword, x_("h")) == 0 ||
			namesame(keyword, x_("x")) == 0)
		{
			if (namesame(keyword, x_("l")) == 0) command = VECTORL; else
				if (namesame(keyword, x_("h")) == 0) command = VECTORH; else
					command = VECTORX;
			for(;;)
			{
				keyword = getkeyword(&pt, x_(" "));
				if (keyword == NOSTRING || *keyword == 0) break;
				par[0] = keyword;
				irsim_newvector(command, 1, par, -1.0);
			}
			continue;
		}

		/* handle commands */
		if (namesame(keyword, x_("clock")) == 0) command = VECTORCLOCK; else
		if (namesame(keyword, x_("vector")) == 0) command = VECTORVECTOR; else
		if (namesame(keyword, x_("c")) == 0) command = VECTORC; else
		if (namesame(keyword, x_("assert")) == 0) command = VECTORASSERT; else
		if (namesame(keyword, x_("stepsize")) == 0) command = VECTORSTEPSIZE; else
		if (namesame(keyword, x_("s")) == 0) command = VECTORS; else
		if (namesame(keyword, x_("!")) == 0) command = VECTOREXCL; else
		if (namesame(keyword, x_("?")) == 0) command = VECTORQUESTION; else
		if (namesame(keyword, x_("set")) == 0) command = VECTORSET; else
		{
			ttyputmsg(_("Unknown IRSIM command on line %ld: %s"), lineno, buf);
			continue;
		}
		count = 0;
		for(;;)
		{
			keyword = getkeyword(&pt, x_(" "));
			if (keyword == NOSTRING || *keyword == 0) break;
			if (count >= irsim_inputparametertotal)
			{
				newtotal = irsim_inputparametertotal * 2;
				if (count >= irsim_inputparametertotal)
					newtotal = count + 5;
				newparamlist = (CHAR **)emalloc(newtotal * (sizeof (CHAR *)), sim_tool->cluster);
				if (newparamlist == 0) return;
				for(i=0; i<irsim_inputparametertotal; i++)
					newparamlist[i] = irsim_inputparameters[i];
				for(i=irsim_inputparametertotal; i<newtotal; i++)
					newparamlist[i] = 0;
				if (irsim_inputparametertotal > 0) efree((CHAR *)irsim_inputparameters);
				irsim_inputparameters = newparamlist;
				irsim_inputparametertotal = newtotal;
			}
			if (irsim_inputparameters[count] != 0) efree((CHAR *)irsim_inputparameters[count]);
			(void)allocstring(&irsim_inputparameters[count], keyword, sim_tool->cluster);
			count++;
		}
		irsim_newvector(command, count, irsim_inputparameters, -1.0);

		/* update the display if this is a "vector" command */
		if (command == VECTORVECTOR)
		{
			/* make the bus */
			bussignals = (INTBIG *)emalloc((count-1) * SIZEOFINTBIG, el_tempcluster);
			if (bussignals == 0) return;

			/* if the bus name already exists, delete it */
			highsiglist = sim_window_findtrace(irsim_inputparameters[0], &highsigcount);
			for(i=0; i<highsigcount; i++)
				sim_window_killtrace(highsiglist[i]);

			j = 0;
			for(i=1; i<count; i++)
			{
				highsiglist = sim_window_findtrace(irsim_inputparameters[i], &highsigcount);
				if (highsigcount == 0)
				{
					ttyputmsg(x_("Vector %s: cannot find signal '%s'"),
						irsim_inputparameters[0], irsim_inputparameters[i]);
				} else
				{
					bussignals[j++] = highsiglist[0];
				}
			}
			if (j > 0)
				(void)sim_window_makebus(j, bussignals, irsim_inputparameters[0]);
			efree((CHAR *)bussignals);
		}
	}
	irsim_playvectors();

	sim_window_gettimeextents(&min, &max);
	sim_window_settimerange(0, min, max);
	sim_window_setmaincursor((max-min)/5.0*2.0+min);
	sim_window_setextensioncursor((max-min)/5.0*3.0+min);
	irsim_UpdateWindow(irsim_cur_delta);

	xclose(vin);
}

/*
 * Routine to save simulation vectors to file "filename" (if zero, prompt for file).
 */
void irsim_savevectorfile(CHAR *filename)
{
	CHAR *truename;
	REGISTER INTBIG i;
	FILE *vout;
	REGISTER SIMVECTOR *sv;
	REGISTER void *infstr;

	if (filename == 0)
	{
		infstr = initinfstr();
		formatinfstr(infstr, x_("%s.cmd"), el_curlib->libname);
		filename = fileselect(_("IRSIM vector file"), sim_filetypeirsimcmd|FILETYPEWRITE,
			returninfstr(infstr));
		if (filename == 0) return;
	}
	vout = xcreate(filename, sim_filetypeirsimcmd, 0, &truename);
	if (vout == 0)
	{
		if (truename != 0) ttyputerr(_("ERROR: Can't create %s"), truename);
		return;
	}

	for(sv = irsim_firstvector; sv != NOSIMVECTOR; sv = sv->nextsimvector)
	{
		if (sv->command == VECTORCOMMENT)
		{
			efprintf(vout, x_("%s\n"), sv->parameters[0]);
			continue;
		}
		efprintf(vout, x_("%s"), irsim_commandname(sv->command));
		for(i=0; i<sv->parcount; i++)
			efprintf(vout, x_(" %s"), sv->parameters[i]);
		efprintf(vout, x_("\n"));
	}

	xclose(vout);
	ttyputmsg(_("Wrote %s"), truename);
}

/*
 * Routine to initialize a list of instance names at the current level.
 */
BOOLEAN irsim_topofinstances(CHAR **c)
{
	irsim_instnetnumber = 0;
	return(TRUE);
}

/*
 * Routine to return the next in the list of instance names at this level.
 */
CHAR *irsim_nextinstance(void)
{
	REGISTER INTBIG i, j, k, len;
	REGISTER CHAR *name, *oname;
	static CHAR retstr[300];

	len = estrlen(irsim_level_prefix);
	for(;;)
	{
		i = irsim_instnetnumber++;
		if (i >= sim_irsimnetnumber) break;
		if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;

		name = sim_irsimnets[i].name;
		if (len > 0 && namesamen(name, irsim_level_prefix, len) != 0)
			continue;

		for(j=len+1; name[j] != 0; j++)
			if (name[j] == '/') break;
		if (name[j] == 0) continue;

		/* new name found: make sure it is unique */
		for(k=0; k<i; k++)
		{
			oname = sim_irsimnets[k].name;
			if (j > 0 && namesamen(oname, name, j+1) == 0)
				break;
		}
		if (k < i) continue;

		name[j] = 0;
		estrcpy(retstr, &name[len]);
		name[j] = '/';
		return(retstr);
	}
	return(0);
}

/*
 * Routine to move down the hierarchy into instance "level" and show
 * signals at that level.
 */
void irsim_level_set(CHAR *level, NODEPROTO *cell)
{
	if (level != 0)
	{
		estrcat(irsim_level_prefix, level);
		estrcat(irsim_level_prefix, x_("/"));
	}

	/* remember saved signals in this cell */
	sim_window_grabcachedsignalsoncell(cell);

	/* ready to add: remove highlighting */
	sim_window_cleartracehighlight();

	/* setup signals for this cell */
	irsim_showcurrentlevel(cell);
	irsim_UpdateWindow(irsim_cur_delta);
	irsim_playvectors();
}

/*
 * Routine to move up one level of hierarchy in the display of signal names.
 */
void irsim_level_up(NODEPROTO *cell)
{
	REGISTER INTBIG len, i;

	len = estrlen(irsim_level_prefix) - 2;
	if (len > 0)
	{
		for(i=len; i>=0; i--)
			if (irsim_level_prefix[i] == '/') break;
		irsim_level_prefix[i+1] = 0;
	}

	/* remember saved signals in this cell */
	sim_window_grabcachedsignalsoncell(cell);

	/* ready to add: remove highlighting */
	sim_window_cleartracehighlight();

	/* setup signals for this cell */
	irsim_showcurrentlevel(cell);
	irsim_UpdateWindow(irsim_cur_delta);
	irsim_playvectors();
}

void irsim_adddisplayedsignal(CHAR *sig)
{
	REGISTER INTBIG i;
	INTSML inistate[1];
	double initime[1];

	/* find the signal */
	for(i=0; i<sim_irsimnetnumber; i++)
	{
		if ((sim_irsimnets[i].flags&IRSIMNETSHOWN) != 0) continue;
		if ((sim_irsimnets[i].flags&IRSIMNETVALID) == 0) continue;
		if (namesame(sig, sim_irsimnets[i].name) == 0) break;
	}
	if (i >= sim_irsimnetnumber) return;

	/* ready to add: remove highlighting */
	sim_window_cleartracehighlight();

	/* create a new trace in the last slot */
	sim_irsimnets[i].signal = sim_window_newtrace(-1, sim_irsimnets[i].name, 0);
	sim_window_addhighlighttrace(sim_irsimnets[i].signal);
	initime[0] = 0.0;   inistate[0] = (LOGIC_X << 8) | GATE_STRENGTH;
	sim_window_setnumframes(sim_window_getnumframes()+1);
	sim_window_loaddigtrace(sim_irsimnets[i].signal, 1, initime, inistate);
	
	irsim_UpdateWindow(irsim_cur_delta);
}

/*
 * Routine that feeds the current signals into the explorer window.
 */
void irsim_reportsignals(WINDOWPART *simwin, void *(*addbranch)(CHAR*, void*),
	void *(*findbranch)(CHAR*, void*), void *(*addleaf)(CHAR*, void*), CHAR *(*nodename)(void*))
{
	REGISTER INTBIG i, j, k, len;
	REGISTER CHAR *pt, *signame, save;
	REGISTER void *curlevel, *nextlevel;

	for(i=0; i<sim_irsimnetnumber; i++)
	{
		signame = sim_irsimnets[i].name;
		len = strlen(signame);
		for(j=len-2; j>0; j--)
			if (signame[j] == '/') break;
		if (j <= 0)
		{
			/* no "." qualifiers */
			(*addleaf)(signame, 0);
		} else
		{
			/* is down the hierarchy */
			signame[j] = 0;

			/* now find the location for this branch */
			curlevel = 0;
			pt = signame;
			for(;;)
			{
				for(k=0; pt[k] != 0; k++) if (pt[k] == '/') break;
				save = pt[k];
				pt[k] = 0;
				nextlevel = (*findbranch)(pt, curlevel);
				if (nextlevel == 0)
					nextlevel = (*addbranch)(pt, curlevel);
				curlevel = nextlevel;
				if (save == 0) break;
				pt[k] = save;
				pt = &pt[k+1];
			}
			(*addleaf)(&signame[j+1], curlevel);
			signame[j] = '/';
		}
	}
}

#if SIMFSDBWRITE != 0
/*************************** FSDB WRITING *******************************/

/*
 * Routine to write history into Novas FSDB file.
 */
void irsim_writeFsdb()
{
	ffwObject *fsdb_obj;
	INTBIG     i;
	CHAR      *p;
	CHAR	   fname[256];
	CHAR       path[1024];

	esnprintf(fname, 256, x_("electric.fsdb"));
	fsdb_obj = ffw_Open(fname, FSDB_FT_EPIC);

	ffw_CreateTreeByHandleScheme(fsdb_obj);
#ifdef PSEC_TIME
	ffw_SetScaleUnit(fsdb_obj, x_("0.01n"));
#else
	ffw_SetScaleUnit(fsdb_obj, x_("0.1n"));
#endif

	/* Tree creation */
	ffw_BeginTree(fsdb_obj);
	for (i = 0; i < sim_irsimaliasnumber; i++)
	{
		p = estrrchr(sim_irsimaliases[i].alias, '/');
		if (p == NULL)
		{
			esnprintf(path, 1024, x_("Top"));
			p = sim_irsimaliases[i].alias;
		} else
		{
			esnprintf(path, 1024, x_("Top/%s"), sim_irsimaliases[i].alias);
			path[p - sim_irsimaliases[i].alias + 4] = 0;
			p++;
		}
		if (NULL == ffw_CreateVarByHandleAndPath(fsdb_obj, FSDB_VT_EPIC_LOGIC, 
				FSDB_VD_IMPLICIT, FSDB_DT_HANDLE_EPIC_DIGITAL, 0, 0,
				sim_irsimaliases[i].name, p, FSDB_BYTES_PER_BIT_1B, path, x_("/")))
		{
			ttyputerr(x_("failed to create a var(%s)"), sim_irsimaliases[i].alias);
			return;
		}
	}
	ffw_EndTree(fsdb_obj);

	/* Value changes */
	irsim_walk_net_index(irsim_writeNodeFsdb, (FILE*)fsdb_obj);

	ffw_Close(fsdb_obj);
}

/*
 * Routine to write node history into Novas FSDB file..
 */
static int irsim_writeNodeFsdb(nptr nd, Ulong ndindx, FILE *arg)
{
	hptr        hp;
	INTBIG      l, h, i;
	fsdbTag64 	time;
	byte_T      value;
	ffwObject  *fsdb_obj = (ffwObject*)arg;

	/* Search node in sim_irsimnodes */
	l = 0;
	h = sim_irsimnetnumber;
	while (l < h)
	{
		int m = (l + h) / 2;
		if (namesame(nd->nname, sim_irsimnets[m].name) <= 0)
			h = m;
		else
			l = m + 1;
	}
	if (l >= sim_irsimnetnumber || namesame(nd->nname, sim_irsimnets[l].name) != 0)
	{
		ttyputmsg(x_("IrsimFsdb: node %s not found"), nd->nname);
		return(1);
	}

	/* Write history in signal-based mode */
	ffw_ResetXCoor(fsdb_obj);
	for (hp = &(nd->head); hp != irsim_last_hist; hp = hp->next)
	{
		if (hp->punt) continue;
		time.L = (fsdbTag32)hp->htime;
		time.H = (fsdbTag32)(hp->htime >> 32);
		switch (hp->val)
		{
			case LOW:
				value = FSDB_BT_VCD_0;
				break;
			case HIGH:
				value = FSDB_BT_VCD_1;
				break;
			default:
				value = FSDB_BT_VCD_X;
		}
		ffw_CreateXCoorByHnL(fsdb_obj, time.H, time.L);
		ffw_CreateVarValueByHandle(fsdb_obj, sim_irsimnets[l].name, &value);
	}
	return(0);
}
#endif
