/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"


void (*irsim_model_table[NMODEL])(nptr) =	/* model dispatch table */
{
	irsim_linear_model, irsim_switch_model
};


int    irsim_model_num = LIN_MODEL;		/* index of model in table */
void (*irsim_model)(nptr) = irsim_linear_model;	/* evaluation model */
int    irsim_sm_stat = NORM_SIM;		/* simulation status */
int    irsim_treport = 0;			/* report format */

static	int	firstcall = 1;	    /* reset when calling init_vdd_gnd */

/* Routine added to initialize globals */
void irsim_init_eval(void)
{
	firstcall = 1;
	irsim_model_num = LIN_MODEL;
	irsim_model = irsim_linear_model;
	irsim_sm_stat = NORM_SIM;
	irsim_treport = 0;
}

/*
 * find transistors with gates of VDD or GND and calculate values for source
 * and drain nodes just in case event driven calculations don't get to them.
 */
static void init_vdd_gnd(void)
{
	irsim_enqueue_input(irsim_VDD_node, HIGH);
	irsim_enqueue_input(irsim_GND_node, LOW);

	firstcall = 0;		/* initialization now taken care of */
}

/*
 * Reset the firstcall flag.  Usually after reading history dump or net state
 */
void irsim_NoInit(void)
{
	firstcall = 0;
}


/*
 * Set the firstcall flags.  Used when moving back to time 0.
 */
void irsim_ReInit(void)
{
	firstcall = 1;
}

/*
 * Print decay event.
 */
static void pr_decay(evptr e)
{
	nptr  n = e->enode;

	ttyputmsg(x_(" @ %.2fns %s: decay %c -> X"),
		d2ns(e->ntime), pnode(n), irsim_vchars[n->npot]);
}


/*
 * Print watched node event.
 */
static void pr_watched(evptr e, nptr n)
{
	int   tmp;
	CHAR buf[200];

	if (n->nflags & INPUT)
	{
		ttyputmsg(x_(" @ %.2fns input %s: -> %c"),
			d2ns(e->ntime), pnode(n), irsim_vchars[e->eval]);
		return;
	}

	tmp = (irsim_debug & DEBUG_EV) ? (REPORT_TAU | REPORT_DELAY) : irsim_treport;

	buf[0] = 0;
	switch(tmp & (REPORT_TAU | REPORT_DELAY))
	{
		case REPORT_TAU:
			esnprintf(buf, 100, x_(" (tau=%.2fns)"), d2ns(e->rtime));
			break;
		case REPORT_DELAY:
			esnprintf(buf, 100, x_(" (delay=%.2fns)"), d2ns(e->delay));
			break;
		default:
			esnprintf(buf, 100, x_(" (tau=%.2fns, delay=%.2fns)"),
				d2ns(e->rtime), d2ns(e->delay));
	}
	ttyputmsg(x_(" @ %.2fns %s: %c -> %c%s"),
		d2ns(e->ntime), pnode(n), irsim_vchars[n->npot], irsim_vchars[e->eval], buf);
}

#ifdef POWER_EST
/*
 * Print capwatched node event.
 */
static void pr_capwatched(evptr e, nptr n)
{
	int   tmp;

	if (irsim_caplogfile == NULL) 
		return;

	if (n->nflags & INPUT)
	{
		efprintf(irsim_caplogfile, x_("%.2f *INPUT* %s -> %c\t"),
			d2ns(e->ntime), pnode(n), irsim_vchars[e->eval]);
		efprintf(irsim_caplogfile, x_("%.2f %.2f %4.3f\n"),
			d2ns(e->rtime), d2ns(e->delay), n->ncap);
		return;
	}

	efprintf(irsim_caplogfile,x_("%.2f\t%s\t%c -> %c\t"),
		d2ns(e->ntime), pnode(n), irsim_vchars[n->npot], irsim_vchars[e->eval]);
	efprintf(irsim_caplogfile, x_("%.2f %.2f %4.3f\n"),
		d2ns(e->rtime), d2ns(e->delay), n->ncap);
	n->toggles++;
	irsim_toggled_cap += n->ncap;    /* When finished, P = toggle_cap * v*v / (2t) */
}


/*
 * Tally cap * trans for giving stepwise power display
 */
static void acc_step_power(nptr n)
{
	if (! (n->nflags & INPUT))
		irsim_step_cap_x_trans += n->ncap;
}
#endif

/*
 * Run through the event list, marking all nodes that need to be evaluated.
 */
static void MarkNodes(evptr evlist)
{
	register nptr   n;
	register tptr   t;
	register lptr   l;
	register evptr  e = evlist;
	long            all_flags = 0;

	do
	{
		n = e->enode;

		all_flags |= n->nflags;

		if (e->type == DECAY_EV && ((irsim_treport & REPORT_DECAY) ||
			(n->nflags & (WATCHED | STOPONCHANGE))))
				pr_decay(e);
		else if (n->nflags & (WATCHED | STOPONCHANGE))
			pr_watched(e, n);
#ifdef POWER_EST
		else if (n->nflags & (POWWATCHED | STOPONCHANGE))
			pr_capwatched(e, n);
		if (irsim_pstep && (n->nflags & (POWWATCHED | STOPONCHANGE)))
			acc_step_power(n);
#endif

		n->npot = e->eval;

		/* Add the new value to the history list (if they differ) */
		if (! (n->nflags & INPUT) && ((short)n->curr->val != n->npot))
			irsim_AddHist(n, n->npot, 0, e->ntime, (long) e->delay,
		(long) e->rtime);

		if (n->awpending != NULL  && n->awpot == n->npot) 
			irsim_evalAssertWhen(n);

#ifdef STATS
		{ extern int ev_hgm; if (ev_hgm) IncHistEvCnt(-1); }
#endif

		/* for each transistor controlled by event node, mark 
		 * source and drain nodes as needing recomputation.
		 *
		 * Added MOSSIMs speed up by first checking if the 
		 * node needs to be rechecked  mh
		 *
		 * Fixed it so nodes with pending events also get
		 * re_evaluated. Kevin Karplus
		 */
		for(l = n->ngate; l != NULL; l = l->next)
		{
#ifdef notdef
			CHAR  oldstate;

			t = l->xtor;
			/*
			 * Some of these optimizations only work right when the
			 * stage at the src/drn contains no loops.  For example,
			 * when a transistor turns off and the src=1 and drn=0,
			 * the transistor may break a previous loop causing the
			 * src or drn to change value.
			 * 
			 * Also, while the state of the src/drn of the transistor
			 * in question may not change state, the current through
			 * their stage is altered; this current change may change
			 * the state of other nodes in the stage, so they better
			 * be avaluated.  These optimizations work best in MOSSIM
			 * since that model does not account for the current through
			 * the stage.	A.S.
			 */
			oldstate = t->state;

			t->state = compute_trans_state(t);
			if ((t->drain->npot == X) || (t->source->npot == X) or
				((t->drain->npot != t->source->npot) &&
				(t->state == ON)) ||
				((t->drain->npot == t->source->npot) &&
				(t->state == OFF)) ||
				((t->state == UNKNOWN) &&
				! (oldstate == OFF &&
				(t->drain->npot == t->source->npot))) ||
				(t->drain->events != NULL) ||
				(t->source->events != NULL))
			{
				if (! (t->source->nflags & INPUT))
					t->source->nflags |= VISITED;
				if (! (t->drain->nflags & INPUT))
					t->drain->nflags |= VISITED;
			}
#else
			t = l->xtor;
#ifdef USER_SUBCKT 
			if (t->ttype != SUBCKT)
			{
				t->state = compute_trans_state(t);
				if (! (t->source->nflags & INPUT))
					t->source->nflags |= VISITED;
				if (! (t->drain->nflags & INPUT))
					t->drain->nflags |= VISITED;
			} 
			else t->drain->nflags |= VISITED;
#else
			t->state = compute_trans_state(t);
			if (! (t->source->nflags & INPUT))
				t->source->nflags |= VISITED;
			if (! (t->drain->nflags & INPUT))
				t->drain->nflags |= VISITED;
#endif
#endif
		}
		free_from_node(e, n);    /* remove to avoid punting this event */
		e = e->flink;
	}
	while(e != NULL);

	/* run thorugh event list again, marking src/drn of input nodes */
	if (all_flags & INPUT)
	{
		for(e = evlist; e != NULL; e = e->flink)
		{
			n = e->enode;

			if ((n->nflags & (INPUT | POWER_RAIL)) != INPUT)
				continue;

			for(l = n->nterm; l != NULL; l = l->next)
			{
				t = l->xtor;
#ifdef USER_SUBCKT 
				if (t->ttype != SUBCKT)
				{
#endif
					if (t->state != OFF)
					{
						register nptr other = other_node(t, n);
						if (!(other->nflags & (INPUT | VISITED)))
							other->nflags |= VISITED;
					}
#ifdef USER_SUBCKT 
				} else
				{
					t->drain->nflags |= VISITED;
				}
#endif
			}
		}
	}
}


static long EvalNodes(evptr evlist)
{
	register tptr   t;
	register lptr   l;
	register nptr   n;
	register evptr  event = evlist;
	long            brk_flag = 0;

	do
	{
		irsim_nevent += 1;		/* advance counter to that of this event */
		n = irsim_cur_node = event->enode;
		n->c.time = event->ntime;	/* set up the cause stuff */
		n->t.cause = event->p.cause;

		irsim_npending -= 1;

		/* now calculate new value for each marked node.  Some nodes marked
		* above may become unmarked by earlier calculations before we get
		* to them in this loop...
		*/

		for(l = n->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
#ifdef USER_SUBCKT 
			if (t->ttype == SUBCKT)
			{
				if (t->drain->nflags & VISITED)
				{
					if (t->subptr->model != (ufun)NULL)
						irsim_subckt_model_C(t);
					else ttyputmsg(x_("Warning: null subckt function"));
				}
			} else 
#endif
			{
				if (t->source->nflags & VISITED)
					(*irsim_model)(t->source);
				if (t->drain->nflags & VISITED)
					(*irsim_model)(t->drain);
			}
		}

		if ((n->nflags & (INPUT | POWER_RAIL)) == INPUT)
		{
			for(l = n->nterm; l != NULL; l = l->next)
			{
				nptr  other;

				t = l->xtor;
				other = other_node(t, n);
				if (other->nflags & VISITED)
					(*irsim_model)(other);
			}
		}

		/* see if we want to halt if this node changes value */
		brk_flag |= n->nflags;

		event = event->flink;
	}
	while(event != NULL);

	return(brk_flag);
}


/*
 * Change the state of the nodes in the given input list to their new value,
 * setting their INPUT flag and enqueueing the event.
 */
static void SetInputs(register iptr  *listp, register int val)
{
	register nptr  n;
	iptr           ip, last;

	for(ip = last = *listp; ip != NULL; ip = ip->next)
	{
		last = ip;

		n = ip->inode;

		n->npot = val;
		n->nflags &= ~INPUT_MASK;
		n->nflags |= INPUT;

		/* enqueue event so consequences are computed. */
		irsim_enqueue_input(n, val);

		if ((int)n->curr->val != val || ! (n->curr->inp))
			irsim_AddHist(n, val, 1, irsim_cur_delta, 0L, 0L);
	}

	if (last)
	{
		last->next = irsim_infree;
		irsim_infree = *listp;
	}
	*listp = NULL;
}


static void MarkNOinputs(void)
{
	register iptr  list;

	for(list = irsim_xinputs; list != NULL; list = list->next)
	{
		list->inode->nflags &= ~(INPUT_MASK | INPUT);
		list->inode->nflags |= VISITED;
	}
}


	/* nodes which are no longer inputs */
static void EvalNOinputs(void)
{
	nptr  n;
	iptr  list, last;

	for(list = last = irsim_xinputs; list != NULL; list = list->next)
	{
		irsim_cur_node = n = list->inode;
		irsim_AddHist(n, (int) n->curr->val, 0, irsim_cur_delta, 0L, 0L);
		if (n->nflags & VISITED)
			(*irsim_model)(n);
		last = list;
	}
	if (last)
	{
		last->next = irsim_infree;
		irsim_infree = irsim_xinputs;
	}
	irsim_xinputs = NULL;
}


int irsim_step(TimeType stop_time)
{
	evptr  evlist;
	long   brk_flag;
	int    ret_code = 0;

	/* look through input lists updating any nodes which just become inputs */

	MarkNOinputs();			/* nodes no longer inputs */
	SetInputs(&irsim_hinputs, HIGH);	/* HIGH inputs */
	SetInputs(&irsim_linputs, LOW);		/* LOW inputs */
	SetInputs(&irsim_uinputs, X);		/* X inputs */

	/* 
	* On the first call to step, make sure transistors with gates
	* of vdd and gnd are set up correctly.  Mark initial inputs first!
	*/
	if (firstcall)
		init_vdd_gnd();

	try_again:
	/* process events until we reach specified stop time or events run out. */
	while((evlist = irsim_get_next_event(stop_time)) != NULL)
	{
		MarkNodes(evlist);
		if (irsim_xinputs) EvalNOinputs();

		brk_flag = EvalNodes(evlist);

		FreeEventList(evlist);	/* return event list to free pool */

		if (stopping(STOPREASONSIMULATE))
			goto done;
		if (brk_flag & (WATCHVECTOR | STOPONCHANGE | STOPVECCHANGE))
		{
			if (brk_flag & (WATCHVECTOR | STOPVECCHANGE))
				irsim_disp_watch_vec(brk_flag);
			if (brk_flag & (STOPONCHANGE | STOPVECCHANGE))
			{
				ret_code = 1;
				goto done;
			}
		}
	}

	if (irsim_xinputs)
	{
		EvalNOinputs();
		goto try_again;
	}

	irsim_cur_delta = stop_time;
done:
	if (irsim_analyzerON)
		irsim_UpdateWindow(irsim_cur_delta);
	return(ret_code);
}


/* table to convert transistor type and gate node value into switch state
 * indexed by irsim_switch_state[transistor-type][gate-node-value].
 */
CHAR  irsim_switch_state[NTTYPES][4] = 
{
	OFF,	UNKNOWN,	UNKNOWN,	ON,		/* NCHAH */
	ON,		UNKNOWN,	UNKNOWN,	OFF,	/* PCHAN */
	WEAK,	WEAK,		WEAK,		WEAK,   /* RESIST */
	WEAK,	WEAK,		WEAK,		WEAK,   /* DEP */
};

int irsim_ComputeTransState(register tptr t)
{
	register nptr  n;
	register tptr  l;
	register int   result;

	switch(BASETYPE(t->ttype))
	{
		case NCHAN:
			result = ON;
			for(l = (tptr) t->gate; l != NULL; l = l->scache.t)
			{
				n = l->gate;
				if (n->npot == LOW)
					return(OFF);
				else if (n->npot == X)
					result = UNKNOWN;
			}
			return(result);

		case PCHAN:
			result = ON;
			for(l = (tptr) t->gate; l != NULL; l = l->scache.t)
			{
				n = l->gate;
				if (n->npot == HIGH)
					return(OFF);
				else if (n->npot == X)
					result = UNKNOWN;
			}
			return(result);

		case DEP:
		case RESIST:
			return(WEAK);

		default:
			ttyputmsg(x_("**** internal error: unrecongized transistor type (0x%x)"),
				BASETYPE(t->ttype));
			return(UNKNOWN);
	}
}

/* VARARGS */
void irsim_error(CHAR *filename, int lineno, CHAR *fmt, ...)
{
	va_list  args;
	CHAR     buf1[100], buf2[ 200];

	var_start(args, fmt);
	(void)esnprintf(buf1, 100, x_("(%s,%d): "), filename, lineno);
	evsnprintf(buf2, 200, fmt, args);
	va_end(args);
	ttyputerr(x_("%s%s"), buf1, buf2);
}
