/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"

#ifdef FAULT_SIM
#  define DoingFault	(irsim_fault_mode != 0)
#else
#  define DoingFault	(0)
#  define hchange		head		/* for the compiler's sake */
#endif


hptr   irsim_freeHist = NULL;	/* list of free history entries */
hptr   irsim_last_hist;			/* pointer to dummy hist-entry that
								 * serves as tail for all nodes */
int    irsim_num_edges = 0;
int    irsim_num_punted = 0;
int    irsim_num_cons_punted = 0;

hptr   irsim_first_model;		/* first model entry */
TimeType irsim_max_time;

static	hptr   curr_model;		/* ptr. to current model entry */
static	CHAR   mem_msg[] = {x_("*** OUT OF MEMORY:Will stop collecting history")};


void irsim_init_hist(void)
{
	static HistEnt  dummy;
	static HistEnt  dummy_model;

	irsim_max_time = (TimeType)MAX_TIME;

	irsim_last_hist = &dummy;
	dummy.next = irsim_last_hist;
	dummy.htime = irsim_max_time;
	dummy.val = X;
	dummy.inp = 1;
	dummy.punt = 0;
	dummy.t.r.delay = dummy.t.r.rtime = 0;

	dummy_model.htime = 0;
	dummy_model.val = irsim_model_num;
	dummy_model.inp = 0;
	dummy_model.punt = 0;
	dummy_model.next = NULL;
	irsim_first_model = curr_model = &dummy_model;

	/* initialize globals */
	irsim_freeHist = NULL;
	irsim_num_edges = 0;
	irsim_num_punted = 0;
	irsim_num_cons_punted = 0;
}


static hptr allocHist(void)
{
	hptr newhist;

	if ((newhist = irsim_freeHist) == NULL)
	{
		if ((newhist = (hptr) irsim_MallocList(sizeof(HistEnt), 0)) == NULL)
		{
			ttyputmsg(mem_msg);
			irsim_sm_stat |= OUT_OF_MEM;
			ttyputmsg(x_("*** can't continue incremetal simulation"));
			return(0);
		}
	}
	irsim_freeHist = newhist->next;
	return(newhist);
}


void irsim_SetFirstHist(nptr node, int value, int inp, TimeType time)
{
	node->head.htime = time;
	node->head.val = value;
	node->head.inp = inp;
	node->head.t.r.rtime = node->head.t.r.delay = 0;
}


/*
 * Add a new entry to the history list.  Update curr to point to this change.
 */
void irsim_AddHist(register nptr node, int value, int inp, TimeType time, long delay, long rtime)
{
	register hptr  newh, curr;

	irsim_num_edges++;

	curr = node->curr;

	if (irsim_sm_stat & OUT_OF_MEM)
		return;

	if (node->nflags & HIST_OFF)
	{ /* Old entries are deleted. Keep only last entry for delay calculation */
		irsim_FreeHistList(node);
		curr = node->curr;
	}

	while(curr->next->punt)		/* skip past any punted events */
		curr = curr->next;

	newh = allocHist();
	if (newh == 0) return;

	newh->next = curr->next;
	newh->htime = time;
	newh->val = value;
	newh->inp = inp;
	newh->punt = 0;
	newh->t.r.delay = (short)delay;
	newh->t.r.rtime = (short)rtime;
	node->curr = curr->next = newh;
}


#define	PuntTime(H)		((TimeType)(H)->htime - (TimeType)(H)->t.p.ptime)


/*
 * Add a punted event to the history list for the node.  Consecutive punted
 * events are kept in punted-order, so that h->ptime < h->next->ptime.
 * Adding a punted event does not change the current pointer, which always
 * points to the last "effective" node change.
 */
void irsim_AddPunted(register nptr node, evptr ev, TimeType tim)
{
	register hptr  newp, h;

	h = node->curr;

	irsim_num_punted++;
	if ((irsim_sm_stat & OUT_OF_MEM) || (node->nflags & HIST_OFF) || DoingFault)
		return;

	newp = allocHist();
	if (newp == 0) return;

	newp->htime = ev->ntime;
	newp->val = ev->eval;
	newp->inp = 0;
	newp->punt = 1;
	newp->t.p.delay = ev->delay;
	newp->t.p.rtime = ev->rtime;
	newp->t.p.ptime = (Uint)((TimeType)newp->htime - tim);

	if (h->next->punt)		/* there are some punted events already */
	{
		irsim_num_cons_punted++;
		do { h = h->next; } while(h->next->punt);
	}

	newp->next = h->next;
	h->next = newp;
}


/*
 * Free up a node's history list
 */
void irsim_FreeHistList(register nptr node)
{
	register hptr  h, next;

	if ((h = node->head.next) == irsim_last_hist)		/* nothing to do */
		return;

	while((next = h->next) != irsim_last_hist)		/* find last entry */
		h = next;

	h->next = irsim_freeHist;				/* link list to free list */
	irsim_freeHist = node->head.next;

	node->head.next = irsim_last_hist;
	node->curr = &(node->head);

	irsim_sm_stat &= ~OUT_OF_MEM;
}



/*
 * Add a new model entry, recording the time of the change.
 */
void irsim_NewModel(int nmodel)
{
	if ((TimeType)curr_model->htime != irsim_cur_delta)
	{
		hptr newh;

		newh = allocHist();
		if (newh == 0) return;

		newh->next = NULL;
		newh->htime = irsim_cur_delta;
		newh->val = nmodel;
		curr_model->next = newh;
		curr_model = newh;
	}
	else
		curr_model->val = nmodel;
}



#define	QTIME(H)		((TimeType)(H)->htime - (TimeType)(H)->t.r.delay)

/*
 * Add a new change to the history list of node 'nd' caused by event 'ev'.
 * Skip past any punted events already in the history and update nd->curr to
 * point to the new change.
 */
void irsim_NewEdge(nptr nd, evptr ev)
{
	register hptr  p, h, newh;

	for(p = nd->curr, h = p->next; h->punt; p = h, h = h->next);

	if (DoingFault)
		p = newh = &nd->hchange;
	else
	{
		newh = allocHist();
		if (newh == 0) return;
	}

	newh->htime = ev->ntime;
	newh->val = ev->eval;
	newh->inp = 0;		/* always true in incremental simulation */
	newh->punt = 0;
	newh->t.r.delay = (short)ev->delay;
	newh->t.r.rtime = ev->rtime;

	p->next = newh;
	newh->next = h;

	nd->curr = newh;		/* newh becomes the current value */
}


/*
 * Delete the next effective change (following nd->curr) from the history.
 * Punted events before the next change (in nd->t.punts) can now be freed.
 * Punted events following the deleted edge are moved to nd->t.punts.
 */
void irsim_DeleteNextEdge(nptr nd)
{
	register hptr  a, b, c;

	if (DoingFault)
	{
		if (nd->t.punts != NULL) ttyputmsg(x_("non-null punts"));

		if (nd->curr != &nd->hchange)
		{
			nd->hchange = *nd->curr;
			nd->curr = &nd->hchange;
		}

		a = nd->curr;
		for(b = a->next; b->punt; a = b, b = b->next);

		nd->hchange.next = b->next;	/* b->next => punt before next edge */
		nd->t.punts = NULL;
		return;
	}

	if ((a = nd->t.punts) != NULL)	/* remove previously punted events */
	{
		for(b = a; b->next != NULL; b = b->next);
		b->next = irsim_freeHist;
		irsim_freeHist = a;
	}

	a = nd->curr;
	for(b = a->next; b->punt; a = b, b = b->next);
	for(c = b->next; c->punt; b = c, c = c->next);
	c = a->next;			/* c => next edge */
	a->next = b->next;
	a = c->next;			/* a => first punted event after c */

	c->next = irsim_freeHist;			/* free the next effective change */
	irsim_freeHist = c;

	if (a->punt)			/* move punted events from hist */
	{
		nd->t.punts = a;
		b->next = NULL;
	}
	else
		nd->t.punts = NULL;
}


#define	NEXTH(H, P)	for((H) = (P)->next; (H)->punt; (H) = (H)->next)

void irsim_FlushHist(TimeType ftime)
{
	register nptr  n;
	register hptr  h, p, head;

	for(n = irsim_GetNodeList(); n != NULL; n = n->n.next)
	{
		head = &(n->head);
		if (head->next == irsim_last_hist || (n->nflags & ALIAS))
			continue;
		p = head;
		NEXTH(h, p);
		while((TimeType)h->htime < ftime)
		{
			p = h;
			NEXTH(h, h);
		}
		head->val = p->val;
		head->htime = p->htime;
		head->inp = p->inp;
		while(p->next != h)
			p = p->next;
		if (head->next != h)
		{
			p->next = irsim_freeHist;
			irsim_freeHist = head->next;
			head->next = h;
		}
		if ((TimeType)n->curr->htime < ftime)
		{
			n->curr = head;
		}
	}
}


int irsim_backToTime(register nptr nd)
{
	register hptr  h, p;

	if (nd->nflags & (ALIAS | MERGED))
		return(0);

	h = &(nd->head);
	NEXTH(p, h);
	while((TimeType)p->htime < irsim_cur_delta)
	{
		h = p;
		NEXTH(p, p);
	}
	nd->curr = h;

	/* queue pending events */
	for(p = h, h = p->next; ; p = h, h = h->next)
	{
		TimeType  qtime;

		if (h->punt)
		{
			if (PuntTime(h) < irsim_cur_delta)	/* already punted, skip it */
				continue;

			qtime = (TimeType)h->htime - (TimeType)h->t.p.delay;	/* pending, enqueue it */
			if (qtime < irsim_cur_delta)
			{
				TimeType tmp = irsim_cur_delta;
				irsim_cur_delta = qtime;
				irsim_enqueue_event(nd, (int) h->val, (long) h->t.p.delay,
					(long) h->t.p.rtime);
				irsim_cur_delta = tmp;
			}
			p->next = h->next;
			h->next = irsim_freeHist;
			irsim_freeHist = h;
			h = p;
		} else
		{
			qtime = QTIME(h);
			if (qtime < irsim_cur_delta)		/* pending, enqueue it */
			{
				TimeType tmp = irsim_cur_delta;
				irsim_cur_delta = qtime;
				irsim_enqueue_event(nd, (int) h->val, (long) h->t.r.delay,
					(long) h->t.r.rtime);
				irsim_cur_delta = tmp;

				p->next = h->next;		/* and free it */
				h->next = irsim_freeHist;
				irsim_freeHist = h;
				h = p;
			}
			else
				break;
		}
	}

	p->next = irsim_last_hist;
	p = h;
	/* p now points to the 1st event in the future (to be deleted) */
	if (p != irsim_last_hist)
	{
		while(h->next != irsim_last_hist)
			h = h->next;
		h->next = irsim_freeHist;
		irsim_freeHist = p;
	}

	h = nd->curr;
	nd->npot = h->val;
	nd->c.time = (TimeType)h->htime;
	if (h->inp)
		nd->nflags |= INPUT;

	if (nd->ngate != NULL)		/* recompute transistor states */
	{
		register lptr  l;
		register tptr  t;

		for(l = nd->ngate; l != NULL; l = l->next)
		{
			t = l->xtor;
			t->state = compute_trans_state(t);
		}
	}
	return(0);
}
