/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/*
 * Incremental net changes are provides by the following commands:
 * 
 * alias node		'==' <node> <node_name>
 * alias node_ref	'=' <node> <node_ref>
 * create new node	'n' <node> <cap> <node_name>
 * eliminate node	'e' <node>
 * eliminate node	'E' <node_name>
 * change node cap	'C' <node> <delta_cap>
 * set node cap		'C' <node> '='<cap>
 * rename node		'r' <node> <node_name>
 * hierarchical rename	'h' <node> <node_name> [<old_name>]
 * connect nodes	'c' <node> <node>
 * break node		'b' <node> <new_node> <cap> <node_name>
 * 
 * add new transistor	'a' <type> <x> <y> <length> <width> <gate> <src> <drn>
 * delete transistor	'd' <x> <y>
 * exchange src/drn	'x' <x> <y>
 * change trans. size	's' <x> <y> <length> <width>
 * change position	'p' <x> <y> <new_x> <new_y>
 * move trans. terminal	'm' <x> <y> <new_gate> <new_src> <new_drn>
 * move trans. to node	'M' <x> <y> <new_node> ['g']['s']['d']
 */

#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"
#include <math.h>    /* **mdg** */

/******** ALIAS MANAGMENT ******/

#define	ALIAS_TBL_SIZE		512
#define	LOG2_TBL_SIZE		9
#define	AliasTblIndex(N)	((N) & (ALIAS_TBL_SIZE - 1))
#define	AliasTblBucket(N)	((N) >> LOG2_TBL_SIZE)
#define	SIZEOF(A)		((int) sizeof(A))


typedef struct
{
	nptr    **bucket;
	int     size;
} Aliases;


static	Aliases	aliases;


static void InitAliasTbl(void)
{
	int  i;

	aliases.size = 20;			/* default initial size */

	aliases.bucket = (nptr **)emalloc(SIZEOF(nptr *) * aliases.size, sim_tool->cluster);
	for(i = 0; i < aliases.size; i++)
		aliases.bucket[i] = NULL;
}


static void FreeAliasTbl(void)
{
	int  i;

	if (aliases.bucket != NULL)
	{
		for(i = 0; i < aliases.size; i++)
		{
			if (aliases.bucket[i] != NULL)
				efree((CHAR *)aliases.bucket[i]);
		}
		efree((CHAR *)aliases.bucket);
	}
	aliases.bucket = NULL;
	aliases.size = 0;
}


static void IncreaseAliasTbl(int i)
{
	Aliases  old;

	old = aliases;

	do { aliases.size *= 2; } while(i >= aliases.size);

	aliases.bucket = (nptr **)emalloc(SIZEOF(nptr *) * aliases.size, sim_tool->cluster);
	for(i = 0; i < old.size; i++) aliases.bucket[i] = old.bucket[i];
	while(i < aliases.size) aliases.bucket[i++] = NULL;
	efree((CHAR *)old.bucket);
}


static void EnterAlias(register int n, nptr nd)
{
	register int   i;
	register nptr  *tbl;

	i = AliasTblBucket(n);
	if (i >= aliases.size)
		IncreaseAliasTbl(i);

	if ((tbl = aliases.bucket[i]) == NULL)
	{
		register nptr  *t, *t_end;

		tbl = (nptr *)emalloc(SIZEOF(nptr) * ALIAS_TBL_SIZE, sim_tool->cluster);
		aliases.bucket[i] = tbl;
		for(t = tbl, t_end = t + ALIAS_TBL_SIZE; t < t_end; *t++ = NULL);
	}

	tbl[AliasTblIndex(n)] = nd;
}


static nptr LookupAlias(register int n)
{
	register int   i;
	register nptr  *t, nd;

	i = AliasTblBucket(n);
	if (i >= aliases.size || (t = aliases.bucket[i]) == NULL)
		return(NULL);
	nd = t[AliasTblIndex(n)];
	while(nd != NULL && (nd->nflags & ALIAS))
		nd = nd->nlink;
	return(nd);
}

/***** END OF ALIAS MANAGMENT ******/


#define LSIZE		2000		/* size for input line buffer */
#define	MAXARGS		20		/* max. # of arguments per line */

#define	streq(A, B)	(estrcmp((A), (B)) == 0)


    /* a capacitance change > CAP_THRESH % will consider the node changed */
#define	CAP_THRESH	(0.05)		/* default is now 5% */


static	int   lineno;			/* current input file line number */
static	CHAR *nu_fname;			/* current input filename */
static	int   num_errors;		/* number of errors in input file */
static	FILE *nu_logf;			/* file for recording net changes */
static	int   context;			/* 0 for old transistor context */
static	int   num_deleted;		/* number of nodes deleted */
static	int   num_cap_set;		/* number of nodes with cap set */
static	nptr  ch_nlist;			/* list of nodes that were changed */
static	tptr  ch_tran;			/* list of changed transistors */
static	nptr  new_GND, new_VDD;
static	int   chg_VDD, chg_GND;
static	void  nu_error(CHAR *fmt, ...);


	/* value stored in tflags, indicating how transistor changed */
#define	T_CH_POS	0x1
#define	T_CHK_TERM	0x2
#define	T_MOV_GATE	0x4
#define	T_MOV_SRC	0x8
#define	T_MOV_DRN	0x10
#define	T_ADDED		(T_CH_POS | T_MOV_GATE | T_MOV_SRC | T_MOV_DRN)

#define	N_CAP_SET	(STIM)


#define	TRAN_CHANGE(T, WHAT)			\
{						\
	if ((T)->tflags == 0)			\
	{						\
		(T)->dcache.t = ch_tran;		\
		(T)->scache.t = ch_tran->scache.t;	\
		ch_tran->scache.t->dcache.t = (T);	\
		ch_tran->scache.t = (T);		\
	}						\
	(T)->tflags |= (WHAT);			\
}


#define	CAP_CHANGE(ND, LIST, GCAP)		\
{						\
	if (! ((ND)->nflags & VISITED))		\
	{						\
		(ND)->n.next = (LIST);			\
		LIST = (ND);				\
		(ND)->nflags |= VISITED;		\
		(ND)->c.cap = (ND)->ncap;		\
	}						\
	(ND)->ncap += (float)(GCAP);			\
}						\


#define	NODE_CHANGE(ND, LIST)			\
{						\
	if (! ((ND)->nflags & VISITED))		\
	{						\
		(ND)->n.next = (LIST);			\
		LIST = (ND);				\
	}						\
	(ND)->nflags |= (VISITED | CHANGED);	\
}						\


#define Error(A)	\
{		\
	nu_error A ;	\
	return;		\
}


static	CHAR	bad_arg_msg[] = {x_("Wrong # of arguments for '%s' expect %s")};
static	CHAR	bad_alias_msg[] = {x_("Illegal alias number (%d)")};
static	CHAR	no_node_msg[] = {x_("can not find node %s")};
static	CHAR	no_alias_msg[] = {x_("Non-existent node alias (%d)")};
static	CHAR	no_trans_msg[] = {x_("can not find transistor @ %d,%d")};


#define	FindNode(ND, ARG) 	\
{				\
	int  Nnum = eatoi(ARG);	\
	if (Nnum < 0)		\
		Error((bad_alias_msg, Nnum)); \
	if (((ND) = LookupAlias(Nnum)) == NULL) \
		Error((no_alias_msg, Nnum)); \
}


#define	ff2pf(C)	((C) * 0.001)		/* ff to pf */


static void alias_node(int ac, CHAR  *av[])
{
	int   nnum;
	nptr  nd;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	nnum = eatoi(av[1]);
	if (nnum < 0) Error((bad_alias_msg, nnum));

	nd = (av[0][1] != '=') ? irsim_FindNode_TxtorPos(av[2]) : irsim_GetNode(av[2]);
	if (nd == NULL)
		Error((no_node_msg, av[2]));

	EnterAlias(nnum, nd);
}


static void new_node(int ac, CHAR  *av[])
{
	nptr    nd;
	int     n;
	double  cap;

	if (ac != 4) Error((bad_arg_msg, av[0], x_("4")));

	cap = ff2pf(eatoi(av[2]));
	n = eatoi(av[1]);
	if (n < 0) Error((bad_alias_msg, n));

	nd = irsim_GetNewNode(av[3]);
	nd->ncap = (float)cap;
	NODE_CHANGE(nd, ch_nlist);

	EnterAlias(n, nd);
}


static void eliminate_node(int ac, CHAR *av[])
{
	nptr  nd;

	if (ac != 2) Error((bad_arg_msg, av[0], x_("2")));

	if (av[0][0] == 'E')
	{
		nd = irsim_find(av[1]);
		if (nd == NULL) Error((no_node_msg, av[1]));
		while(nd->nflags & ALIAS)
			nd = nd->nlink;
	} else
	{
		FindNode(nd, av[1]);
	}

	if (nd->nflags & POWER_RAIL)		/* never delete Vdd or Gnd */
		return;
	irsim_n_delete(nd);
	NODE_CHANGE(nd, ch_nlist);
	nd->nflags |= DELETED;
	irsim_nnodes--;
	num_deleted += 1;
}


static void change_cap(int ac, CHAR *av[])
{
	nptr    nd;
	double  cap;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	FindNode(nd, av[1]);

	if (av[2][0] == '=')
	{
		cap = ff2pf(eatoi(&av[2][1]));
		CAP_CHANGE(nd, ch_nlist, 0);
		nd->ncap = (float)cap;
		nd->nflags |= N_CAP_SET;
		num_cap_set ++;
	} else
	{
		cap = ff2pf(eatoi(av[2]));
		CAP_CHANGE(nd, ch_nlist, cap);
	}
}


static void ChangeNodeName(nptr nd, CHAR  *str)
{
	int  len;

	if (nd->nflags & POWER_RAIL)
	{
		if (namesame(str, nd->nname) == 0)
			return;
		else if (nd == irsim_GND_node)
			chg_GND = TRUE;
		else
			chg_VDD = TRUE;
	}

	len = estrlen(str) + 1;

	irsim_n_delete(nd);

	nd->nname = (CHAR *)emalloc(len * SIZEOFCHAR, sim_tool->cluster);
	memcpy(nd->nname, str, len);

	CAP_CHANGE(nd, ch_nlist, 0);

	if (namesame(str, x_("Vdd")))
		new_VDD = nd;
	else if (namesame(str, x_("Gnd")))
		new_GND = nd;
}


static void rename_node(int ac, CHAR *av[])
{
	nptr  nd;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	FindNode(nd, av[1]);

	if (! streq(nd->nname, av[2]))
		ChangeNodeName(nd, av[2]);
}


/*
 * Return TRUE if 'name1' is "better" than 'name2', FALSE otherwise.
 */
static int BestNodeName(CHAR *name1, CHAR *name2)
{
	register int nslashes1, nslashes2;
	register CHAR *np1, *np2;

	for(nslashes1 = 0, np1 = name1; *np1;)
	{
		if (*np1++ == '/')  nslashes1++;
	}

	for(nslashes2 = 0, np2 = name2; *np2;)
	{
		if (*np2++ == '/')  nslashes2++;
	}

	--np1;
	--np2;

	/* chose label over generated name */
	if (*np1 != '#' && *np2 == '#')	return(TRUE);
	if (*np1 == '#' && *np2 != '#')	return(FALSE);

	/* either both are labels or generated names */
	/* check pathname components */
	if (nslashes1 < nslashes2)		return(TRUE);
	if (nslashes1 > nslashes2)		return(FALSE);

	/* same # of pathname components; check length */
	if (np1 - name1 < np2 - name2)	return(TRUE);
	if (np1 - name1 > np2 - name2)	return(FALSE);

	/* same # of pathname components; same length; use lex ordering */
	return((estrcmp(name1, name2) > 0) ? TRUE : FALSE);
}



static void hier_rename_node(int ac, CHAR *av[])
{
	nptr  nd;

	if (ac < 3 || ac > 4)
		Error((bad_arg_msg, av[0], x_("3-4")));

	FindNode(nd, av[1]);

	if (ac == 3)
	{
		if (BestNodeName(av[2], nd->nname))
			ChangeNodeName(nd, av[2]);
	}
	else if (streq(nd->nname, av[3]))
		ChangeNodeName(nd, av[2]);
}


static void connect_nodes(int ac, CHAR *av[])
{
	int            t_terms;
	register nptr  nd1, nd2;
	register lptr  l;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	FindNode(nd1, av[1]);
	FindNode(nd2, av[2]);

	if (nd1 == nd2)
		return;

	if ((nd1->nflags & nd2->nflags) & POWER_RAIL)
		Error((x_("Warning: attempt to short Vdd & Gnd")));

	if ((nd2->nflags & POWER_RAIL) || BestNodeName(nd2->nname, nd1->nname))
		SWAP_NODES(nd1, nd2);

	if (nd2->ngate != NULL)
	{
		register lptr  last;

		for(l = nd2->ngate; l != NULL; l = l->next)
		{
			l->xtor->gate = nd1;
			last = l;
		}
		last->next = nd1->ngate;
		nd1->ngate = nd2->ngate;
	}

	if (nd2->nterm != NULL)
	{
		register lptr  last;
		register tptr  t;

		for(l = nd1->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			TRAN_CHANGE(t, T_CHK_TERM);
		}
		for(l = nd2->nterm; l != NULL; l = l->next)
		{
			t = l->xtor;
			if (t->source == nd2) t->source = nd1;
			if (t->drain == nd2) t->drain = nd1;
			TRAN_CHANGE(t, T_CHK_TERM);
			last = l;
		}

		if (nd1->nflags & POWER_RAIL)
			last->next = irsim_freeLinks,	irsim_freeLinks = nd2->nterm;
		else
			last->next = nd1->nterm,	nd1->nterm = nd2->nterm;

		t_terms = TRUE;
	}
	else
		t_terms = FALSE;

	for(l = irsim_on_trans; l != NULL; l = l->next)
	{
		if (l->xtor->gate == nd2) l->xtor->gate = nd1;
	}

	if (irsim_tcap->scache.t != irsim_tcap)
	{
		register tptr  t;

		for(t = irsim_tcap->scache.t; t != irsim_tcap; t = t->scache.t)
		{
			if (t->gate == nd2)	t->gate = nd1;
			if (t->source == nd2)	t->source = nd1;
			if (t->drain == nd2)	t->drain = nd1;
		}
	}

	irsim_n_delete(nd2);

	nd2->ngate = nd2->nterm = NULL;
	nd2->nlink = nd1;
	nd2->nflags = (ALIAS | DELETED);
	num_deleted += 1;
	irsim_nnodes--;

	NODE_CHANGE(nd2, ch_nlist);

	CAP_CHANGE(nd1, ch_nlist, nd2->ncap);
	nd2->ncap = 0.0;
	if (t_terms)
		nd1->nflags |= CHANGED;
}


static void break_node(int ac, CHAR *av[])
{
	int     nnum;
	double  cap;
	nptr    nd, new_nd;

	if (ac != 5) Error((bad_arg_msg, av[0], x_("5")));

	FindNode(nd, av[1]);
	CAP_CHANGE(nd, ch_nlist, 0.0);

	nnum = eatoi(av[2]);
	if (nnum < 0) Error((bad_alias_msg, nnum));

	cap = ff2pf(eatoi(av[3]));
	new_nd = irsim_GetNewNode(av[4]);
	new_nd->ncap = (float)cap;
	EnterAlias(nnum, new_nd);
	NODE_CHANGE(new_nd, ch_nlist);
}


static void add_new_trans(int ac, CHAR *av[])
{
	struct Trans  tran;
	tptr          t;
	long          length, width;
	double        cap;

	if (ac != 9) Error((bad_arg_msg, av[0], x_("9")));

	switch(av[1][0])
	{
		case 'n':	tran.ttype = NCHAN;	break;
		case 'p':	tran.ttype = PCHAN;	break;
		case 'd':	tran.ttype = DEP;	break;
		default:
			nu_error(x_("unknown transistor type (%s)"), av[1]);
			return;
	}

	tran.x.pos = eatoi(av[2]);
	tran.y.pos = eatoi(av[3]);
	length = (long)(eatof(av[4]) * irsim_LAMBDACM);
	width = (long)(eatof(av[5]) * irsim_LAMBDACM);

	FindNode(tran.gate, av[6]);
	FindNode(tran.source, av[7]);
	FindNode(tran.drain, av[8]);

	tran.tflags = tran.n_par = 0;
	tran.state = (tran.ttype & ALWAYSON) ? WEAK : UNKNOWN;

	irsim_ntrans[tran.ttype]++;

	NEW_TRANS(t);
	*t = tran;
	t->r = irsim_requiv((int) tran.ttype, width, length);

	t->tlink = t;

	cap = length * width * irsim_CTGA;
	CAP_CHANGE(t->gate, ch_nlist, cap);

	if ((irsim_config_flags & TDIFFCAP) && (cap = irsim_CTDW * width + irsim_CTDE) != 0.0)
	{
		CAP_CHANGE(t->source, ch_nlist, cap);
		CAP_CHANGE(t->drain, ch_nlist, cap);
	}
	if (t->source != t->drain)
	{
		NODE_CHANGE(t->source, ch_nlist);
		NODE_CHANGE(t->drain, ch_nlist);
	}

	TRAN_CHANGE(t, T_ADDED);
}


#define	FindTxtor(T, X_STR, Y_STR, X, Y)	\
{						\
	X = eatoi(X_STR);				\
	Y = eatoi(Y_STR);				\
	if (((T) = irsim_FindTxtorPos(X, Y)) == NULL)	\
	{						\
		nu_error(no_trans_msg, X, Y); 	\
		return;					\
	}						\
}


static void delete_trans(int ac, CHAR *av[])
{
	tptr    t;
	long    x, y;
	double  cap, gcap;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	FindTxtor(t, av[1], av[2], x, y);

	irsim_DeleteTxtorPos(t);

	gcap = -(t->r->width * t->r->length * irsim_CTGA);
	cap = (irsim_config_flags & TDIFFCAP) ? -(irsim_CTDW * t->r->width + irsim_CTDE) : 0.0;

	if (t->ttype & TCAP)
	{
		if (cap != 0.0)
		{
			CAP_CHANGE(t->source, ch_nlist, cap);
			CAP_CHANGE(t->drain, ch_nlist, cap);
		}
		UNLINK_TCAP(t);
	} else
	{
		if (t->ttype & STACKED) irsim_DestroyStack(t->dcache.t);
		if (t->ttype & ORLIST) irsim_UnParallelTrans(t);

		if (t->ttype & ALWAYSON)
		{
			DISCONNECT(irsim_on_trans, t);
		} else
		{
			DISCONNECT(t->gate->ngate, t);
		}
		DISCONNECT(t->source->nterm, t);
		DISCONNECT(t->drain->nterm, t);

		NODE_CHANGE(t->source, ch_nlist);
		NODE_CHANGE(t->drain, ch_nlist);
		if (cap != 0.0)
		{
			t->source->ncap += (float)cap;
			t->drain->ncap += (float)cap;
		}
		if (t->tflags != 0)	/* it is in changed list -> remove it */
		{
			t->dcache.t->scache.t = t->scache.t;
			t->scache.t->dcache.t = t->dcache.t;
		}
	}
	CAP_CHANGE(t->gate, ch_nlist, gcap);
	irsim_ntrans[BASETYPE(t->ttype)] -= 1;
	FREE_TRANS(t);
}


static void exchange_terms(int ac, CHAR *av[])
{
	tptr  t;
	long  x, y;

	if (ac != 3) Error((bad_arg_msg, av[0], x_("3")));

	FindTxtor(t, av[1], av[2], x, y);

	SWAP_NODES(t->source, t->drain);
}


static void change_tsize(int ac, CHAR *av[])
{
	tptr     t;
	long     x, y, length, width;
	double   cap;
	Resists  *old_r;

	if (ac != 5) Error((bad_arg_msg, av[0], x_("5")));

	FindTxtor(t, av[1], av[2], x, y);

	length = (long)(eatof(av[3]) * (float)irsim_LAMBDACM);
	width = (long)(eatof(av[4]) * (float)irsim_LAMBDACM);

	cap = (width * length - t->r->length * t->r->width) * irsim_CTGA;
	CAP_CHANGE(t->gate, ch_nlist, cap);

	if (irsim_config_flags & TDIFFCAP)
	{
		cap = irsim_CTDW * (width - t->r->width) + irsim_CTDE;
		CAP_CHANGE(t->source, ch_nlist, cap);
		CAP_CHANGE(t->drain, ch_nlist, cap);
	}

	old_r = t->r;
	t->r = irsim_requiv((int) t->ttype, width, length);

	if (old_r->dynlow != t->r->dynlow || old_r->dynhigh != t->r->dynhigh)
	{
		NODE_CHANGE(t->source, ch_nlist);
		NODE_CHANGE(t->drain, ch_nlist);
	}
}


static void change_tposition(int ac, CHAR *av[])
{
	tptr  t;
	long  x, y;

	if (ac != 5) Error((bad_arg_msg, av[0], x_("5")));

	FindTxtor(t, av[1], av[2], x, y);

	irsim_DeleteTxtorPos(t);

	t->x.pos = eatoi(av[1]);
	t->y.pos = eatoi(av[2]);

	TRAN_CHANGE(t, T_CH_POS);
}


static void move_trans_terms(int ac, CHAR *av[])
{
	long    x, y, was_tcap = FALSE;
	tptr    t;
	nptr    gate, src, drn;
	double  cap;

#define IsDot(STR)	((STR)[0] == '.' && (STR)[1] == '\0')

	if (ac != 6) Error((bad_arg_msg, av[0], x_("6")));

	FindTxtor(t, av[1], av[2], x, y);

	if (IsDot(av[3]))    gate = t->gate;	else FindNode(gate, av[3]);
	if (IsDot(av[4]))    src = t->source;	else FindNode(src, av[4]);
	if (IsDot(av[5]))    drn = t->drain;	else FindNode(drn, av[5]);

	if (gate == t->gate && src == t->source && drn == t->drain)
		return;

	if (t->ttype & STACKED)	irsim_DestroyStack(t->dcache.t);
	if (t->ttype & ORLIST)	irsim_UnParallelTrans(t);
	if (t->ttype & TCAP)
	{
		UNLINK_TCAP(t);
		was_tcap = TRUE;
	}

	if (gate != t->gate)
	{
		if (gate->nflags & MERGED)
		{
			irsim_DestroyStack(gate->t.tran);
			NODE_CHANGE(gate, ch_nlist);
		}
		cap = (t->r->width * t->r->length) * irsim_CTGA;
		CAP_CHANGE(gate, ch_nlist, cap);
		CAP_CHANGE(t->gate, ch_nlist, -cap);

		if (! (t->ttype & ALWAYSON))
		{
			DISCONNECT(t->gate->ngate, t);
			TRAN_CHANGE(t, T_MOV_GATE);
		}
		t->gate = gate;
	}

	cap = (irsim_config_flags & TDIFFCAP) ? t->r->width * irsim_CTDW + irsim_CTDE : 0.0;

	if (t->source != src)
	{
		if (was_tcap)
		{
			CAP_CHANGE(t->source, ch_nlist, -cap);
		} else
		{
			DISCONNECT(t->source->nterm, t);
			NODE_CHANGE(t->source, ch_nlist);
			t->source->ncap -= (float)cap;
		}
		src->ncap += (float)cap;
		t->source = src;
		TRAN_CHANGE(t, T_MOV_SRC);
	}

	if (t->drain != drn)
	{
		if (was_tcap)
		{
			CAP_CHANGE(t->drain, ch_nlist, -cap);
		} else
		{
			DISCONNECT(t->drain->nterm, t);
			NODE_CHANGE(t->drain, ch_nlist);
			t->drain->ncap -= (float)cap;
		}
		drn->ncap += (float)cap;
		t->drain = drn;
		TRAN_CHANGE(t, T_MOV_DRN);
	}

	NODE_CHANGE(src, ch_nlist);
	NODE_CHANGE(drn, ch_nlist);
}


static void SwitchContext(void)
{
	register tptr  t;

	if (context)
		return;

	context = 1;

	for(t = ch_tran->scache.t; t != ch_tran; t = t->scache.t)
	{
		if (t->tflags & T_CH_POS)
		{
			irsim_EnterPos(t, TRUE);
			t->tflags &= ~T_CH_POS;
		}
	}
}


static void move_trans_to_node(int ac, CHAR *av[])
{
	long    x, y, was_tcap = FALSE;
	tptr    t;
	nptr    nd, gate, src, drn;
	CHAR    *s;
	double  cap;

	SwitchContext();

	if (ac != 5) Error((bad_arg_msg, av[0], x_("5")));

	FindTxtor(t, av[1], av[2], x, y);

	FindNode(nd, av[3]);

	gate = t->gate;
	src = t->source;
	drn = t->drain;
	for(s = av[4]; *s != '\0'; s++)
	{
		if (*s == 'g')		gate = nd;
		else if (*s == 's')	src = nd;
		else if (*s == 'd')	drn = nd;
		else Error((x_("Unknown terminal '%c'"), *s));
	}

	if (gate == t->gate && src == t->source && drn == t->drain)
		return;

	if (t->ttype & STACKED)	irsim_DestroyStack(t->dcache.t);
	if (t->ttype & ORLIST)	irsim_UnParallelTrans(t);
	if (t->ttype & TCAP)
	{
		UNLINK_TCAP(t);
		was_tcap = TRUE;
	}

	if (gate != t->gate)
	{
		if (t->ttype & ALWAYSON )
		{
			DISCONNECT(irsim_on_trans, t);
		} else
		{
			DISCONNECT(t->gate->ngate, t);
		}
		t->gate = gate;
		TRAN_CHANGE(t, T_MOV_GATE);
		cap = (t->r->width * t->r->length) * irsim_CTGA;
		gate->ncap += (float)cap;
	}

	cap = (irsim_config_flags & TDIFFCAP) ? t->r->width * irsim_CTDW + irsim_CTDE : 0.0;

	if (t->source != src)
	{
		if (! was_tcap)
		{
			DISCONNECT(t->source->nterm, t);
		}
		t->source = src;
		src->ncap += (float)cap;
		TRAN_CHANGE(t, T_MOV_SRC);
	}

	if (t->drain != drn)
	{
		if (! was_tcap)
		{
			DISCONNECT(t->drain->nterm, t);
		}
		t->drain = drn;
		drn->ncap += (float)cap;
		TRAN_CHANGE(t, T_MOV_DRN);
	}
}


static void change_thresh(int ac, CHAR *av[])
{
	nptr   nd;
	float  vlow, vhigh;

	if (ac != 4) Error((bad_arg_msg, av[0], x_("4")));
	nd = irsim_find(av[1]);
	while(nd->nflags & ALIAS) nd = nd->nlink;
	if (nd == NULL) Error((no_node_msg, av[1]));
	vlow = (float)eatof(av[2]);
	vhigh = (float)eatof(av[3]);
	if (nd->vlow != vlow || nd->vhigh != vhigh)
	{
		nd->vlow = vlow; nd->vhigh = vhigh;
		NODE_CHANGE(nd, ch_nlist);
	}
}


static void ndelay(int ac, CHAR *av[])
{
	nptr   nd;
	long   tplh, tphl;

	if (ac != 4) Error((bad_arg_msg, av[0], x_("4")));
	nd = irsim_find(av[1]);
	while(nd->nflags & ALIAS) nd = nd->nlink;
	if (nd == NULL) Error((no_node_msg, av[1]));
	tplh = (long)ns2d(eatof(av[2]));
	tphl = (long)ns2d(eatof(av[3]));

	if (! (nd->nflags & USERDELAY) || nd->tplh != tplh || nd->tphl != tphl)
	{
		nd->tplh = (short)tplh; nd->tphl = (short)tphl;
		NODE_CHANGE(nd, ch_nlist);
	}
}


/*
 * parse input line into tokens, filling up carg and setting targc
 */
static int parse_line(CHAR *line, CHAR **carg)
{
	register CHAR  ch;
	register int   targc;

	targc = 0;
	while(ch = *line++)
	{
		if (ch <= ' ')
			continue;
		targc++;
		*carg++ = line - 1;
		while(*line > ' ')
			line++;
		if (*line)
			*line++ = '\0';
	}
	*carg = 0;
	return(targc);
}


static void input_changes(FILE  *fin)
{
	CHAR  line[LSIZE];
	CHAR  *targv[MAXARGS];
	int   targc;

	irsim_VDD_node->nflags |= VISITED;	/* never add these to 'change' list */
	irsim_GND_node->nflags |= VISITED;
	new_GND = new_VDD = NULL;
	chg_GND = chg_VDD = 0;

	InitAliasTbl();

	while(!xfgets(line, LSIZE, fin))
	{
		lineno++;
		if (nu_logf)
			(void)efputs(line, nu_logf);
		targc = parse_line(line, targv);
		if (targc == 0)
			continue;

		switch(targv[0][0])
		{
			case '=': alias_node(targc, targv);			break;
			case 'n': new_node(targc, targv);			break;
			case 'e': eliminate_node(targc, targv);		break;
			case 'E': eliminate_node(targc, targv);		break;
			case 'C': change_cap(targc, targv);			break;
			case 'r': rename_node(targc, targv);		break;
			case 'h': hier_rename_node(targc, targv);	break;
			case 'c': connect_nodes(targc, targv);		break;
			case 'b': break_node(targc, targv);			break;
			case 'a': add_new_trans(targc, targv);		break;
			case 'd': delete_trans(targc, targv);		break;
			case 'x': exchange_terms(targc, targv);		break;
			case 's': change_tsize(targc, targv);		break;
			case 'p': change_tposition(targc, targv);	break;
			case 'm': move_trans_terms(targc, targv);	break;
			case 'M': move_trans_to_node(targc, targv);	break;
			case 't': change_thresh(targc, targv);		break;
			case 'D': ndelay(targc, targv);				break;
			case '|':									break;
			default:  nu_error(x_("Unrecognized command (%s)"), targv[0]);
		}
	}

	(void)fclose(fin);
	if (nu_logf)
		(void)fclose(nu_logf);

	FreeAliasTbl();

	irsim_VDD_node->nflags &= ~(VISITED | CHANGED);	/* restore flag */
	irsim_GND_node->nflags &= ~(VISITED | CHANGED);
}


/*
 * Mark any alias that points to a deleted node as deleted.
 */
static int mark_del_alias(nptr n)
{
	register nptr  nd;
	register int   has_del = FALSE;

	if (n->nflags & DELETED)
		return(0);

	for(nd = n; nd->nflags & ALIAS; nd = nd->nlink)
		if (nd->nflags & DELETED)
			has_del = TRUE;

	if (has_del)		/* deleted node(s) in alias list, resolve */
		n->nlink = nd;

	if (nd->nflags & DELETED)	/* points to a deleted node, delete as well */
	{
		n->nflags |= DELETED;
		n->nflags &= ~ALIAS;
		NODE_CHANGE(n, ch_nlist);
		irsim_naliases--;
	}
	return(0);
}


static nptr rm_nodes(void)
{
	register nptr  *nlast, node;
	nptr           nlist;

	if (num_deleted == 0)
		return(ch_nlist);

	if (irsim_naliases > 0)
		irsim_walk_net(mark_del_alias);

	irsim_rm_del_from_lists();

	nlist = NULL;
	nlast = &nlist;

	for(node = ch_nlist; node != NULL; node = node->n.next)
	{
		if (! (node->nflags & DELETED))
		{
			*nlast = node;
			nlast = &node->n.next;
			continue;
		}

		while(node->events)		/* remove all pending events */
			irsim_free_event(node->events);

		if (IsInList(node->nflags))
			irsim_idelete(node, irsim_listTbl[INPUT_NUM(node->nflags)]);

		irsim_FreeHistList(node);

		if (node->hnext != node)
			irsim_n_delete(node);

		node->nlink = irsim_freeNodes;	/* return it to free pool */
		irsim_freeNodes = node;
	}
	*nlast = NULL;

	return(nlist);
}


static void conn_ch_trans(tptr tlist)
{
	register tptr  t;
	tptr           tnext;

	for(t = tlist->scache.t; t != tlist; t = tnext)
	{
		tnext = t->scache.t;

		if (t->tflags & T_CH_POS)
			irsim_EnterPos(t, TRUE);

		if (t->tflags & (T_MOV_GATE | T_MOV_SRC | T_MOV_DRN))
		{
			nptr  src = t->source, drn = t->drain;

			if (src == drn || (src->nflags & drn->nflags & POWER_RAIL))
			{
				if (! (t->tflags & T_MOV_SRC))
					DISCONNECT(src->nterm, t);
				if (! (t->tflags & T_MOV_DRN))
					DISCONNECT(drn->nterm, t);
				if (t->ttype & ALWAYSON)
					DISCONNECT(irsim_on_trans, t)
				else if (t->tflags & T_MOV_GATE)
					DISCONNECT(t->gate->ngate, t);

				t->ttype |= TCAP;
				LINK_TCAP(t);
			} else
			{
				if (t->tflags & T_MOV_GATE)
				{
					if (t->ttype & ALWAYSON)
						CONNECT(irsim_on_trans, t)
					else
						CONNECT(t->gate->ngate, t);
				}
				if ((t->tflags & T_MOV_SRC) && ! (src->nflags & POWER_RAIL))
					CONNECT(src->nterm, t);
				if ((t->tflags & T_MOV_DRN) && ! (drn->nflags & POWER_RAIL))
					CONNECT(drn->nterm, t);
			}
		} else if (t->tflags & T_CHK_TERM)
		{
			nptr  src = t->source, drn = t->drain;

			if (src == drn || (src->nflags & drn->nflags & POWER_RAIL))
			{
				DISCONNECT(src->nterm, t);
				DISCONNECT(drn->nterm, t);
				if (t->ttype & ALWAYSON)
					DISCONNECT(irsim_on_trans, t)
				else
					DISCONNECT(t->gate->ngate, t);

				t->ttype |= TCAP;
				LINK_TCAP(t);
			}
		}
		t->tflags = 0;
	}
}


static int chk_power(void)
{
	if (new_GND != NULL || chg_GND != 0)
	{
		ttyputerr(x_("Error: changed Gnd, cannot deal with this yet!"));
		return(FALSE);
	}

	if (new_VDD != NULL || chg_VDD != 0)
	{
		ttyputerr(x_("Error: changed Vdd, cannot deal with this yet!"));
		return(FALSE);
	}
	return(TRUE);
}


static void add_tran_cap(tptr t)
{
	if (t->gate->nflags & N_CAP_SET)
		t->gate->ncap += (float)(t->r->width * t->r->length * irsim_CTGA);

	if (irsim_config_flags & TDIFFCAP)
	{
		if (t->source->nflags & N_CAP_SET)
			t->source->ncap += (float)(t->r->width * irsim_CTDW + irsim_CTDE);
		if (t->drain->nflags & N_CAP_SET)
			t->drain->ncap += (float)(t->r->width * irsim_CTDW + irsim_CTDE);
	}
}


#define CHK_CAP(ND)	if ((ND)->ncap < MIN_CAP) (ND)->ncap = (float)MIN_CAP
#define	ABS(X)	(((X) < 0) ? -(X) : (X))


static iptr changed_nodes(void)
{
	register nptr  n;
	iptr           ilist = NULL;

	for(n = ch_nlist; n != NULL; n = n->n.next)
	{
		CHK_CAP(n);

		if (n->hnext == n)
			irsim_n_insert(n);

		n->nflags &= ~N_CAP_SET;

		if (! (n->nflags & CHANGED))
		{
			double deltac = ABS(n->c.cap - n->ncap);
			if (deltac < CAP_THRESH * n->c.cap)
			{
				n->c.time = (TimeType)n->curr->htime;
				continue;
			}
			n->nflags |= CHANGED;
		}

		if (n->nflags & MERGED)
		{
			register nptr  stack;

			stack = n->t.tran->source;
			if (! (stack->nflags & (CHANGED | POWER_RAIL)))
			{
				irsim_iinsert(stack, &ilist);
				stack->nflags |= CHANGED;
			}
			stack = n->t.tran->drain;
			if (! (stack->nflags & (CHANGED | POWER_RAIL)))
			{
				irsim_iinsert(stack, &ilist);
				stack->nflags |= CHANGED;
			}
			n->nflags &= ~CHANGED;
			continue;
		}

		if (n->nflags & POWER_RAIL)
			continue;

		irsim_iinsert(n, &ilist);
	}
	return(ilist);
}


iptr irsim_rd_changes(CHAR *fname, CHAR *logname)
{
	FILE          *fin;
	CHAR          *dummy;
	iptr          ilist = NULL;
	struct Trans  dummyt;

	ch_tran = dummyt.scache.t = dummyt.dcache.t = &dummyt;
	ch_nlist = NULL;
	context = num_cap_set = num_deleted = num_errors = lineno = 0;
	nu_fname = fname;

	fin = xopen(fname, el_filetypetext, 0, &dummy);
	if (fin == NULL)
	{
		ttyputmsg(x_("can not open '%s' for net changes"), fname);
		return(NULL);
	}

	if (logname == NULL)
	{
		nu_logf = NULL;
	} else
	{
		nu_logf = xopen(logname, el_filetypetext|FILETYPEAPPEND, 0, &dummy);
		if (nu_logf == NULL)
		{
			ttyputmsg(x_("warning: can't open logfile %s"), logname);
		} else
		{
			time_t  ltime = time(0);
			(void)efprintf(nu_logf, x_("| changes @ %s"), ectime(&ltime));
		}
	}

	if (irsim_analyzerON)	irsim_StopAnalyzer();

	input_changes(fin);

	if (!chk_power()) return(NULL);
	ch_nlist = rm_nodes();
	conn_ch_trans(ch_tran);
	if (num_cap_set) irsim_walk_trans(add_tran_cap);
	irsim_make_parallel(ch_nlist);
	irsim_make_stacks(ch_nlist);

	irsim_pTotalNodes();
	irsim_pTotalTxtors();
	irsim_pParallelTxtors();
	irsim_pStackedTxtors();

	ilist = changed_nodes();

	if (irsim_analyzerON)	irsim_RestartAnalyzer(irsim_sim_time0, irsim_sim_time0, FALSE);

	if (num_errors != 0)
		ttyputmsg(x_("%s contains %d errors"), fname, num_errors,
			(nu_logf != NULL) ? x_(" listed in logfile") : x_(""));

	if (egetenv(x_("RSIM_CHANGED")) != NULL)
	{
		iptr  ip;

		if (ilist != NULL) ttyputmsg(x_("changed nodes:"));
		for(ip = ilist; ip != NULL; ip = ip->next)
			ttyputmsg(x_("  %s"), pnode(ip->inode));
	}

	return(ilist);
}

static void nu_error(CHAR *fmt, ...)
{
	va_list  args;
	CHAR     *errstr = x_("| error");
	CHAR msg[200];

	var_start(args, fmt);
	if (nu_logf != NULL)
	{
		(void)efprintf(nu_logf, x_("%s:%s, line %d: "), errstr, nu_fname, lineno);
		(void)evfprintf(nu_logf, fmt, args);
	} else
	{
		errstr++;
		evsnprintf(msg, 200, fmt, args);
		ttyputmsg(x_("%s:%s, line %d: %s"), errstr, nu_fname, lineno, msg);
	}
	va_end(args);
	num_errors += 1;
}
