/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/*
 * The routines in this file handle network input (from sim files).
 * The input routine "irsim_rd_network" replaces the work formerly done by presim.
 * This version differs from the former in the following:
 *  1. voltage drops across transistors are ignored (assumes transistors
 *     driven by nodes with voltage drops have the same resistance as those
 *     driven by a full swing).
 *  2. static power calculations not performed (only useful for nmos anyhow).
 */


#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"


#define LSIZE		2000		/* size for input line buffer */
#define	MAXARGS		50		/* max. # of arguments per line */


nptr   irsim_VDD_node;		/* power supply nodes */
nptr   irsim_GND_node;

lptr   irsim_on_trans;		/* always on transistors */

int    irsim_nnodes;			/* number of actual nodes */
int    irsim_naliases;		/* number of aliased nodes */
int    irsim_ntrans[NTTYPES];	/* number of txtors indexed by type */

lptr   irsim_freeLinks = NULL;	/* free list of Tlist structs */
tptr   irsim_freeTrans = NULL;	/* free list of Transistor structs */
nptr   irsim_freeNodes = NULL;	/* free list of Node structs */

tptr   irsim_tcap = NULL;		/* list of capacitor-transistors */

static	int    lineno;			/* current input file line number */
static	CHAR   *simfname;		/* current input filename */
static	int    nerrs = 0;		/* # of erros found in sim file */
static	int    isBinFile;		/* TRUE binary file, FALSE sim file */

static	tptr   rd_tlist;		/* list of transistors just read */


#define	WARNING		0
#define	WARN		1
#define	MAX_ERRS	20
#define	FATAL		(MAX_ERRS + 1)


#define	MIT	0
#define LBL	1
#define	SU	2

static short int simFormat = MIT, offsLBL = 0 ;


static	CHAR    bad_argc_msg[] = {x_("Wrong number of args for '%c' (%d)")};

#define	BAD_ARGC(CMD, ARGC, ARGV)			\
{							\
	irsim_error(simfname, lineno, bad_argc_msg, CMD, ARGC);	\
	PrArgs(ARGC, ARGV);				\
	(void)CheckErrs(WARN);					\
	return;						\
}

static void irsim_freetransistor(tptr t);


/*
 * Routine to free all memory allocated in this module.
 */
void irsim_freesimmemory(void)
{
	register tptr  t;
	lptr l;

	while (irsim_on_trans != 0)
	{
		l = irsim_on_trans;
		irsim_on_trans = l->next;
		efree((CHAR *)l);
	}
	while (rd_tlist != 0)
	{
		t = rd_tlist;
		rd_tlist = t->scache.t;
		irsim_freetransistor(t);
	}
/*	if (irsim_tcap != 0) irsim_freetransistor(t); */
	while (irsim_freeTrans != 0)
	{
		t = irsim_freeTrans;
		irsim_freeTrans = (tptr)t->gate;
		efree((CHAR *)t);
	}
	while (irsim_freeLinks != 0)
	{
		l = irsim_freeLinks;
		irsim_freeLinks = l->next;
		efree((CHAR *)l);
	}
}

void irsim_freetransistor(tptr t)
{
	nptr gate, src, drn;
	lptr l;

	for(gate = t->gate; gate != 0; gate = gate->nlink)
	{
		if ((gate->nflags & ALIAS) != 0) continue;
		while (gate->ngate != 0)
		{
			l = gate->ngate;
			gate->ngate = l->next;
			efree((CHAR *)l);
		}
	}
	for(src = t->source; src != 0; src = src->nlink)
	{
		if ((src->nflags & ALIAS) != 0) continue;
		while (src->nterm != 0)
		{
			l = src->nterm;
			src->nterm = l->next;
			efree((CHAR *)l);
		}
	}
	for(drn = t->drain; drn != 0; drn = drn->nlink)
	{
		if ((drn->nflags & ALIAS) != 0) continue;
		while (drn->nterm != 0)
		{
			l = drn->nterm;
			drn->nterm = l->next;
			efree((CHAR *)l);
		}
	}
	efree((CHAR *)t);
}

static void PrArgs(int argc, CHAR *argv[])
{
	while(argc-- != 0)
		ttyputerr(x_("%s "), *argv++);
}


/*
 * Returns TRUE if there have been too many errors and the activity should be stopped.
 */
static int CheckErrs(int n)
{
	nerrs += n;
	if (nerrs > MAX_ERRS)
	{
		if (n != FATAL)
			ttyputerr(x_("Too many errors in sim file <%s>"),
				simfname);
		return(TRUE);
	}
	return(FALSE);
}



/*
 * Traverse the transistor list and add the node connection-list.  We have
 * to be careful with ALIASed nodes.  Note that transistors with source/drain
 * connected VDD and GND nodes are not linked.
 */
static nptr connect_txtors(void)
{
	register tptr  t, tnext;
	register nptr  gate, src, drn, nd_list;
	register int   type;
	register long  visited;

	visited = VISITED;
	nd_list = NULL;

	for(t = rd_tlist; t != NULL; t = tnext)
	{
		tnext = t->scache.t;
		for(gate = t->gate; gate->nflags & ALIAS; gate = gate->nlink);
		for(src = t->source; src->nflags & ALIAS; src = src->nlink);
		for(drn = t->drain; drn->nflags & ALIAS; drn = drn->nlink);

		t->gate = gate;
		t->source = src;
		t->drain = drn;

		type = t->ttype;
		t->state = (type & ALWAYSON) ? WEAK : UNKNOWN;
		t->tflags = 0;

		irsim_ntrans[type]++;
		if (src == drn || (src->nflags & drn->nflags & POWER_RAIL))
		{
			t->ttype |= TCAP;		/* transistor is just a capacitor */
			LINK_TCAP(t);
		} else
		{
			/* do not connect gate if ALWAYSON since they do not matter */
			if (t->ttype & ALWAYSON)
			{
				CONNECT(irsim_on_trans, t);
			} else
			{
				CONNECT(t->gate->ngate, t);
			}

			if (! (src->nflags & POWER_RAIL))
			{
				CONNECT(src->nterm, t);
				LINK_TO_LIST(src, nd_list, visited);
			}
			if (! (drn->nflags & POWER_RAIL))
			{
				CONNECT(drn->nterm, t);
				LINK_TO_LIST(drn, nd_list, visited);
			}
		}
	}

	return(nd_list);
}


/*
 * node area and perimeter info (N sim command).
 */
static void node_info(int targc, CHAR *targv[])
{
	register nptr  n;

	if (targc != 8)
		BAD_ARGC('N', targc, targv);

	n = irsim_GetNode(targv[1]);

	n->ncap += (float)(eatof(targv[4]) * (irsim_CMA * irsim_LAMBDA2) +
		eatof(targv[5]) * (irsim_CPA * irsim_LAMBDA2) +
		eatof(targv[6]) * (irsim_CDA * irsim_LAMBDA2) +
		eatof(targv[7]) * 2.0f * (irsim_CDP * irsim_LAMBDA));
}


/*
 * new format node area and perimeter info (M sim command).
 */
static void nnode_info(int targc, CHAR *targv[])
{
	register nptr  n;

	if (targc != 14)
		BAD_ARGC('M', targc, targv);

	n = irsim_GetNode(targv[1]);

	n->ncap += (float)(eatof(targv[4]) * (irsim_CM2A * irsim_LAMBDA2) +
		eatof(targv[5]) * 2.0 * (irsim_CM2P * irsim_LAMBDA) +
		eatof(targv[6]) * (irsim_CMA * irsim_LAMBDA2) +
		eatof(targv[7]) * 2.0 * (irsim_CMP * irsim_LAMBDA) +
		eatof(targv[8]) * (irsim_CPA * irsim_LAMBDA2) +
		eatof(targv[9]) * 2.0 * (irsim_CPP * irsim_LAMBDA) +
		eatof(targv[10]) * (irsim_CDA * irsim_LAMBDA) +
		eatof(targv[11]) * 2.0 * (irsim_CDP * irsim_LAMBDA) +
		eatof(targv[12]) * (irsim_CPDA * irsim_LAMBDA2) +
		eatof(targv[13]) * 2.0 * (irsim_CPDP * irsim_LAMBDA));
}

#ifdef USER_SUBCKT 
/*
 * new subckt with (num_inputs, inputs...., outputs...., subckt_name) HMD
 */
static int newsubckt(int targc, CHAR *targv[])
{
	double fabs(), eatof();
	int i, n, ninputs, noutputs, ubleng, dleng;
	long width, length;
	ufun model, init;
	register tptr t;
	tptr t2; lptr l;
	/* tptr sc_list = (tptr)NULL; */
	CHAR *sname, *out_name_Ub, *out_name_D;
	int subckt_out_counter = 0;
	nptr *nodes;
	lptr diodes = (lptr)NULL, d;
	uptr usrData;
	int  inst;

	ninputs   = eatoi(targv[1]);
	noutputs  = (targc - 3 - ninputs)/2;
	sname     = (CHAR *)emalloc((estrlen(targv[targc-1])+1) * SIZEOFCHAR, sim_tool->cluster);
	ubleng = estrlen(targv[targc-1])+20;
	out_name_Ub = (CHAR *)emalloc(ubleng * SIZEOFCHAR, sim_tool->cluster);
	dleng  = estrlen(targv[targc-1])+20;
	out_name_D  = (CHAR *)emalloc(dleng * SIZEOFCHAR, sim_tool->cluster);
	estrcpy(sname, targv[targc-1]);
	model     = irsim_subckt_instantiate(sname, &init, &inst);
	if (model == (ufun)NULL)
	{
		ttyputmsg(x_("Error: subckt `%s' not defined"),sname);
		return(FALSE);
	}
	targc -= 2; targv += 2;
	nodes = (nptr *)emalloc((ninputs + 2*noutputs) * sizeof(nptr), sim_tool->cluster);

	ttyputmsg(x_("defining new blackbox `%s' instance %d #i:%d #o:%d"),
	sname,inst, ninputs, noutputs);

	if (ninputs <= 0)
	{
		ttyputmsg(x_("\terror 0 inputs"));
		return(FALSE);
	}
	if (noutputs < 0)
	{
		ttyputmsg(x_("\terror < 0 outputs"));
		return(FALSE);
	}
	if (targc < ninputs + 1)
	{
		ttyputmsg(x_("\terror wrong # args %d"),targc);
		return(FALSE);
	} 
	if (init)
		usrData = (uptr)(*init)();

	for (i = 0; i<ninputs; i++)
	{
		/* create new transistor */
		float ct ;

		NEW_TRANS(t);		
		NEW_SUBCKT(t->subptr);
		t->ttype = SUBCKT; 
		t->gate = irsim_GetNode(targv[i]);
		t->subptr->nodes = nodes; 
		t->source = irsim_VDD_node; 
		t->drain = t->gate;
		/* link it to the list */
		t->scache.t = rd_tlist; rd_tlist = t; 
		/* t->dcache.t = sc_list; sc_list = t; */
		t->r = irsim_requiv(RESIST, 2 , 1000000*irsim_LAMBDACM);
		t->subptr->model = model; t->subptr->sname = sname;

		/* All this can be referenced by a pointer to a subckt structure */
		NEW_LINK(d); d->xtor = t; d->next = diodes;
		t->subptr->ndiode = d; diodes = d;
		t->subptr->nins = ninputs; t->subptr->nouts = noutputs; 
		nodes[i] = t->gate;
		t->subptr->udata = usrData;
	} 
	for(l = t->subptr->ndiode; l != NULL; l = l->next)
	{
		t2 = l->xtor;
		t2->subptr->ndiode = diodes;
	}

	for (n=i ; i<targc-1; i+=2, n++)
	{
		int rt;
		/* defining two transistors for each output: */

		esnprintf(out_name_Ub, ubleng, x_("%sUb_%d_%d"), sname, inst, subckt_out_counter);
		esnprintf(out_name_D, dleng, x_("%sD_%d_%d"), sname, inst, subckt_out_counter++);
		if (irsim_find(out_name_Ub) || irsim_find(out_name_D))
		{
			ttyputmsg(x_("Error %s or %s already exists"),
				out_name_Ub, out_name_D);
			return(FALSE);
		}
		rt = (int) fabs(eatof(targv[i+1]));

		NEW_TRANS(t);	/* Add pmos output driver */
		t->ttype    = PCHAN;
		t->gate     = irsim_GetNode(out_name_Ub);
		nodes[n]    = t->gate;
		t->drain    = irsim_VDD_node;
		t->source   = irsim_GetNode(targv[i]);
		t->scache.t = rd_tlist; /* link it to the list */
		rd_tlist    = t;
		/* t->dcache.t = sc_list; sc_list = t; */
		t->drain->ncap += 0;
		t->r        = irsim_requiv(RESIST, 2 , rt*irsim_LAMBDACM);

		NEW_TRANS(t);	/* Add nmos output driver */
		t->ttype    = NCHAN;
		t->gate     = irsim_GetNode(out_name_D);
		t->drain    = irsim_GetNode(targv[i]);
		nodes[n+noutputs]    = t->gate;
		t->source   = irsim_GND_node;
		t->scache.t = rd_tlist; /* link it to the list */
		rd_tlist    = t;
		/* t->dcache.t = sc_list; sc_list = t; */
		t->source->ncap += 0;
		t->r        = irsim_requiv(RESIST, 2 , rt*irsim_LAMBDACM);
	} /* for loop for outputs */
	free(out_name_D); free(out_name_Ub);
	return(TRUE);
}
#endif

static	int    AP_error = FALSE;

/* parse g=...,A_NUM,P_NUM attributes */
static int parseAttr(CHAR *str, unsigned int *a, unsigned int *p)
{
	int l;
	CHAR *s;

	if ((l=estrlen(str)) <= 2)
	{
		*a = 0 ; 
		*p = 0;
		return FALSE;
	}

	for (s = str+l*sizeof(CHAR) ; *s != 'A' && s != str ; s--) ;
		if (esscanf(s,x_("A_%d,P_%d"), a, p) != 2) 
			if (esscanf(s,x_("a_%d,p_%d"), a, p) != 2)
	{
		irsim_error(simfname, lineno, x_("Bad area/perimeter attributes"));
		return FALSE;
	}
	return TRUE;
}


/*
 * new transistor.  Implant specifies type.
 * AreaPos specifies the argument number that contains the area (if any).
 */
static void newtrans(int implant, int targc, CHAR *targv[])
{
	nptr           gate, src, drn;
	long           x, y, width, length;
	double         cap;
	register tptr  t;
	unsigned int fetHasAP = FALSE ;
	unsigned int asrc, adrn, psrc, pdrn;

	if (implant == RESIST)
	{
		if (targc != 4)
			BAD_ARGC('r', targc, targv);

		gate = irsim_VDD_node;
		src = irsim_GetNode(targv[1]);
		drn = irsim_GetNode(targv[2]);

		length = (long)(eatof(targv[3]) * irsim_LAMBDACM);
		width = 0;
	}
	else
	{
		if (targc < 4+offsLBL || targc > 11+offsLBL)
			BAD_ARGC(targv[0][0], targc, targv);

		gate = irsim_GetNode(targv[1]);
		src = irsim_GetNode(targv[2]);
		drn = irsim_GetNode(targv[3]);

		if (targc > 5+offsLBL)
		{
			length = (long)(eatof(targv[4+offsLBL]) * irsim_LAMBDACM);
			width = (long)(eatof(targv[5+offsLBL]) * irsim_LAMBDACM);
			if (width <= 0 || length <= 0)
			{
				irsim_error(simfname, lineno,
					x_("Bad transistor width=%g or length=%g"), width, length);
				return;
			}
			if (targc > 7+offsLBL)
			{
				x = eatoi(targv[6+offsLBL]);
				y = eatoi(targv[7+offsLBL]);
			}
		} else
			width = length = 2 * irsim_LAMBDACM;

		cap = length * width * irsim_CTGA;
	}

	NEW_TRANS(t);			/* create new transistor */

	t->ttype = (UCHAR1)implant;
	t->gate = gate;
	t->source = src;
	t->drain = drn;

	if (targc > 7+offsLBL)
	{
		t->x.pos = x;
		t->y.pos = y;
		irsim_EnterPos(t, TRUE);		/* Enter transistor position */
		if (simFormat == SU && targc >= 9)
		{
			/* parse area and perimeter */
			register int i ;

			for (fetHasAP = TRUE, i = 8 ; i < targc ; i++)
			{
				if (targv[i][0] == 's')
					fetHasAP = fetHasAP && parseAttr(targv[i], &asrc, &psrc);
				else if (targv[i][0] == 'd')
					fetHasAP = fetHasAP && parseAttr(targv[i], &adrn, &pdrn);
			}
		}
	} else
	{
		irsim_EnterPos(t, FALSE);		/* Enter transistor position */
		if (simFormat == SU && ! AP_error)
		{
			irsim_error( simfname, lineno, x_("no area/perim S/D attributes on fet"));
			AP_error = TRUE;
		}
	}

	t->scache.t = rd_tlist;		/* link it to the list */
	rd_tlist = t;

	t->r = irsim_requiv(implant, width, length);

	/* update node capacitances  */
	gate->ncap += (float)cap;

	if (simFormat == SU)
	{
		double capsrc = 0, capdrn =0 ;

		if (fetHasAP)
		{
			if (implant == PCHAN)
			{
				capsrc = asrc * irsim_LAMBDA * irsim_LAMBDA * irsim_CPDA + psrc * irsim_LAMBDA * irsim_CPDP;
				capdrn = adrn * irsim_LAMBDA * irsim_LAMBDA * irsim_CPDA + pdrn * irsim_LAMBDA * irsim_CPDP;
			} 
			else if (implant == NCHAN || implant == DEP)
			{
				capsrc = asrc * irsim_LAMBDA * irsim_LAMBDA * irsim_CDA + psrc * irsim_LAMBDA * irsim_CDP;
				capdrn = adrn * irsim_LAMBDA * irsim_LAMBDA * irsim_CDA + pdrn * irsim_LAMBDA * irsim_CDP;
			} 
		} else if (! AP_error )
		{
			ttyputmsg(x_("Warning: Junction capacitances might be incorrect"));
			AP_error = TRUE;
		}
		src->ncap += (float)capsrc;
		drn->ncap += (float)capdrn;
	} else if (irsim_config_flags & TDIFFCAP) 
	{
		if (implant == PCHAN)
		{
			cap = irsim_CPTDW * width + irsim_CPTDE;
		}
		else if (implant == NCHAN || implant == DEP)
			cap = irsim_CTDW * width + irsim_CTDE;
		else
			cap = 0.0 ;
		src->ncap += (float)cap;
		drn->ncap += (float)cap;
	}
}

/*
 * accept a bunch of aliases for a node (= sim command).
 */
static void alias(int targc, CHAR *targv[])
{
	register nptr  n, m;
	register int   i;

	if (targc < 3)
		BAD_ARGC('=', targc, targv);

	n = irsim_GetNode(targv[1]);

	for(i = 2; i < targc; i++)
	{
		m = irsim_GetNode(targv[i]);
		if (m == n)
			continue;

		if (m->nflags & POWER_RAIL)
			SWAP_NODES(m, n);

		if (m->nflags & POWER_RAIL)
		{
			irsim_error(simfname, lineno, x_("Can't alias the power supplies"));
			continue;
		}

		n->ncap += m->ncap;

		m->nlink = n;
		m->nflags |= ALIAS;
		m->ncap = 0.0;
		irsim_nnodes--;
		irsim_naliases++;
	}
}


/*
 * node threshold voltages (t sim command).
 */
static void nthresh(int targc, CHAR *targv[])
{
	register nptr  n;

	if (targc != 4)
		BAD_ARGC('t', targc, targv);

	n = irsim_GetNode(targv[1]);
	n->vlow = (float)eatof(targv[2]);
	n->vhigh = (float)eatof(targv[3]);
}


/*
 * User delay for a node (D sim command).
 */
static void ndelay(int targc, CHAR *targv[])
{
	register nptr  n;

	if (targc != 4)
		BAD_ARGC('D', targc, targv);

	n = irsim_GetNode(targv[1]);
	n->nflags |= USERDELAY;
	n->tplh = (short)ns2d(eatof(targv[2]));
	n->tphl = (short)ns2d(eatof(targv[3]));
}


/*
 * add capacitance to a node (c sim command).
 */
static void ncap(int targc, CHAR *targv[])
{
	register nptr n, m;
	float         cap;

	if (targc == 3)
	{
		n = irsim_GetNode(targv[1]);
		n->ncap += (float)eatof(targv[2]);
	}
	else if (targc == 4)		/* two terminal caps	*/
	{
		cap = (float)(eatof(targv[3]) / 1000);		/* ff to pf conversion */
		n = irsim_GetNode(targv[1]);
		m = irsim_GetNode(targv[2]);
		if (n != m)			/* add cap to both nodes */
		{
			if (m != irsim_GND_node)	m->ncap += cap;
			if (n != irsim_GND_node)	n->ncap += cap;
		}
		else if (n == irsim_GND_node)	/* same node, only GND makes sense */
		n->ncap += cap;
	}
	else
		BAD_ARGC('c', targc, targv);
}


/*
 * parse input line into tokens, filling up carg and setting targc
 */
static int parse_line(CHAR *line, CHAR **carg)
{
	register CHAR  ch;
	register int   targc;

	targc = 0;
	while(ch = *line++)
	{
		if (ch <= ' ')
			continue;
		targc++;
		*carg++ = line - 1;
		while(*line > ' ')
			line++;
		if (*line)
			*line++ = '\0';
	}
	*carg = 0;
	return(targc);
}


static	int    R_error = FALSE;
static	int    A_error = FALSE;

static int input_sim(CHAR *simfile)
{
	FILE  *fin;
	CHAR  line[LSIZE], *dummy;
	CHAR  *targv[MAXARGS];	/* tokens on current command line */
	int   targc;		/* number of args on command line */
	long  offset;		/* state of previously opened file */
	int   olineno;

	fin = xopen(simfile, el_filetypetext, 0, &dummy);
	if (fin == NULL)
	{
		ttyputmsg(x_("cannot open '%s' for sim input"), simfile);
		return(FALSE);
	}
	simfname = simfile;
	lineno = 0;

	while(!xfgets(line, LSIZE, fin))
	{
		lineno++;
		targc = parse_line(line, targv);
		if (targv[0] == NULL) continue;
		switch(targv[0][0])
		{
			case '@':
				if (targc != 2)
				{
					irsim_error(simfname, lineno, bad_argc_msg, '@', targc);
					if (CheckErrs(WARN)) return(FALSE);
					break;
				}
				offset = ftell(fin);
				olineno = lineno;
				xclose(fin);
				(void)input_sim(targv[1]);
				fin = xopen(simfile, el_filetypetext, 0, &dummy);
				if (fin == NULL)
				{
					irsim_error(simfname, lineno, x_("can't re-open sim file '%s'"),
						simfile);
					(void)CheckErrs(WARN);
					return(FALSE);
				}
				(void)fseek(fin, offset, 0);
				simfname = simfile;
				lineno = olineno;
				break;

			case '|':
				if (lineno > 1) break;
				if (targc >= 2)
				{
					double lmbd = eatof(targv[2])/100.0;

					if (lmbd != irsim_LAMBDA)
					{
						ttyputerr(x_("WARNING: sim file lambda (%.2lfu) != config lambda (%.2lfu), using config lambda"),
							lmbd, irsim_LAMBDA);
					}
				}
				if (targc >= 6)
				{
					if (!namesame(targv[6], x_("MIT"))) simFormat = MIT ;
					else if (!namesame(targv[6], x_("LBL"))) simFormat = LBL ;
					else if (!namesame(targv[6], x_("SU"))) simFormat = SU ;
					else
					{
						ttyputerr(x_("Unknown format '%s' assuming MIT"), targv[6]);
						simFormat = MIT ;
					}
					offsLBL = (simFormat == LBL) ? 1 : 0 ;/* skip bulk */
					if (simFormat == SU)
					{
						if (irsim_CDA == 0.0 || irsim_CDP == 0.0 || irsim_CPDA == 0.0 || irsim_CPDP == 0.0)
							ttyputmsg(x_("Warning: SU format and area/perim cap values are zero"));
						else
							ttyputmsg(x_("info: SU format --> using S/D attrs to compute junction parasitics"));
					}
				} 
				break;
			case 'e':
			case 'n':
				newtrans(NCHAN, targc, targv);
				break;
			case 'p':
				newtrans(PCHAN, targc, targv);
				break;
			case 'd':
				newtrans(DEP, targc, targv);
				break;
			case 'r':
				newtrans(RESIST, targc, targv);
				break;
#ifdef	USER_SUBCKT 
			case 'x':
				if (newsubckt(targc, targv) == FALSE) return(NULL);
				break;
#endif
			case 'N':
				node_info(targc, targv);
				break;
			case 'M':
				nnode_info(targc, targv);
				break;
			case 'c':
			case 'C':
				ncap(targc, targv);
				break;
			case '=':
				alias(targc, targv);
				break;
			case 't':
				nthresh(targc, targv);
				break;
			case 'D': 
				ndelay(targc, targv);
				break;
			case 'R': 
				if (! R_error)	/* only warn about this 1 time */
				{
					ttyputmsg(x_("%s: Ignoring lumped-resistance ('R' construct)"),
						simfname);
					R_error = TRUE;
				}
				break;
			case 'A':
				if (! A_error)	/* only warn about this 1 time */
				{
					ttyputmsg(x_("%s: Ignoring attribute-line ('A' construct)"),
						simfname);
					A_error = TRUE;
				}
				break;
			case '<': 
				if (lineno == 1 && irsim_rd_netfile(fin, line))
				{
					(void)fclose(fin);
					return(TRUE);
				}
				/* fall through if irsim_rd_netfile returns FALSE */
			default:
				irsim_error(simfname, lineno, x_("Unrecognized input line (%s)"),
					targv[0]);
				if (CheckErrs(WARN)) return(FALSE);
		}
	}
	xclose(fin);
	ttyputmsg(x_("Loaded circuit, lambda=%.2lfu format=%s"), irsim_LAMBDA, 
		simFormat == MIT ? x_("MIT") : (simFormat == LBL ? x_("LBL") : x_("SU")));
	return(FALSE);
}

static void init_counts(void)
{
	register int  i;

	for(i = 0; i < NTTYPES; i++)
		irsim_ntrans[i] = 0;
	irsim_nnodes = irsim_naliases = 0;
}


int irsim_rd_network(CHAR *simfile)
{
	rd_tlist = NULL;
	irsim_init_hash();
#ifdef	USER_SUBCKT
	irsim_init_subs();
#endif
	init_counts();
	irsim_init_listTbl();
	irsim_init_eval();
	irsim_init_faultsim_module();
	irsim_init_incsim();
	irsim_init_newrstep();
	irsim_init_rsim();
	irsim_init_sched();
	irsim_init_stack();
	irsim_init_tpos();

	irsim_VDD_node = irsim_GetNode(x_("Vdd"));
	irsim_VDD_node->npot = HIGH;
	irsim_VDD_node->nflags |= (INPUT | POWER_RAIL);
	irsim_VDD_node->head.inp = 1;
	irsim_VDD_node->head.val = HIGH;
	irsim_VDD_node->head.punt = 0;
	irsim_VDD_node->head.htime = 0;
	irsim_VDD_node->head.t.r.rtime = irsim_VDD_node->head.t.r.delay = 0;
	irsim_VDD_node->head.next = irsim_last_hist;
	irsim_VDD_node->curr = &(irsim_VDD_node->head);

	irsim_GND_node = irsim_GetNode(x_("Gnd"));
	irsim_GND_node->npot = LOW;
	irsim_GND_node->nflags |= (INPUT | POWER_RAIL);
	irsim_GND_node->head.inp = 1;
	irsim_GND_node->head.val = LOW;
	irsim_GND_node->head.punt = 0;
	irsim_GND_node->head.htime = 0;
	irsim_GND_node->head.t.r.rtime = irsim_GND_node->head.t.r.delay = 0;
	irsim_GND_node->head.next = irsim_last_hist;
	irsim_GND_node->curr = &(irsim_GND_node->head);

	NEW_TRANS(irsim_tcap);
	irsim_tcap->source = 0;
	irsim_tcap->drain = 0;
	irsim_tcap->scache.t = irsim_tcap->dcache.t = irsim_tcap;
	irsim_tcap->x.pos = 0;

	nerrs = 0;

	isBinFile = input_sim(simfile);
	if (nerrs > 0)
		return(1);
	return 0;
}


void irsim_pTotalNodes(void)
{
	REGISTER void *infstr;

	infstr = initinfstr();
	formatinfstr(infstr, x_("%d nodes"), irsim_nnodes);
	if (irsim_naliases != 0)
		formatinfstr(infstr, x_(", %d aliases"), irsim_naliases);
	ttyputmsg(x_("%s"), returninfstr(infstr));
}


void irsim_pTotalTxtors(void)
{
	int  i;
	REGISTER void *infstr;

	infstr = initinfstr();
	addstringtoinfstr(infstr, x_("transistors:"));
	for(i = 0; i < NTTYPES; i++)
	{
		if (irsim_ntrans[i] == 0) continue;
		formatinfstr(infstr, x_(" %s=%d"), irsim_ttype[i], irsim_ntrans[i]);
	}
	if (irsim_tcap->x.pos != 0)
		formatinfstr(infstr, x_(" shorted=%d"), irsim_tcap->x.pos);
	ttyputmsg(x_("%s"), returninfstr(infstr));
}


void irsim_ConnectNetwork(void)
{
	nptr  ndlist;

	irsim_pTotalNodes();

	ndlist = (isBinFile) ? irsim_bin_connect_txtors() : connect_txtors();

	irsim_make_parallel(ndlist);
	irsim_make_stacks(ndlist);

	irsim_pTotalTxtors();
	irsim_pParallelTxtors();
	irsim_pStackedTxtors();
}
